#!/usr/bin/env python

'''
fit best estimate of magnetometer offsets, diagonals, off-diagonals, cmot and scaling using WMM target
'''

from MAVProxy.modules.lib import wx_processguard
from MAVProxy.modules.lib.wx_loader import wx

import sys, time, os, math, copy

from pymavlink import mavutil
from pymavlink import mavextra
from pymavlink.rotmat import Vector3
from pymavlink.rotmat import Matrix3
from MAVProxy.modules.lib import grapher
from MAVProxy.modules.lib.multiproc_util import MPDataLogChildTask

import matplotlib
import matplotlib.pyplot as pyplot
import numpy
import datetime

earth_field = None
declination = None
margs = None
mag_idx = ''

class Correction:
    def __init__(self):
        self.offsets = Vector3(0.0, 0.0, 0.0)
        self.diag = Vector3(1.0, 1.0, 1.0)
        self.offdiag = Vector3(0.0, 0.0, 0.0)
        self.cmot = Vector3(0.0, 0.0, 0.0)
        self.scaling = 1.0

    def show_parms(self):
        print("COMPASS_OFS%s_X %d" % (mag_idx, int(self.offsets.x)))
        print("COMPASS_OFS%s_Y %d" % (mag_idx, int(self.offsets.y)))
        print("COMPASS_OFS%s_Z %d" % (mag_idx, int(self.offsets.z)))
        print("COMPASS_DIA%s_X %.3f" % (mag_idx, self.diag.x))
        print("COMPASS_DIA%s_Y %.3f" % (mag_idx, self.diag.y))
        print("COMPASS_DIA%s_Z %.3f" % (mag_idx, self.diag.z))
        print("COMPASS_ODI%s_X %.3f" % (mag_idx, self.offdiag.x))
        print("COMPASS_ODI%s_Y %.3f" % (mag_idx, self.offdiag.y))
        print("COMPASS_ODI%s_Z %.3f" % (mag_idx, self.offdiag.z))
        print("COMPASS_MOT%s_X %.3f" % (mag_idx, self.cmot.x))
        print("COMPASS_MOT%s_Y %.3f" % (mag_idx, self.cmot.y))
        print("COMPASS_MOT%s_Z %.3f" % (mag_idx, self.cmot.z))
        print("COMPASS_SCALE%s %.2f" % (mag_idx, self.scaling))
        if margs['CMOT']:
            print("COMPASS_MOTCT 2")

def correct(MAG, BAT, c):
    '''correct a mag sample, returning a Vector3'''
    mag = Vector3(MAG.MagX, MAG.MagY, MAG.MagZ)

    # add the given offsets
    mag += c.offsets

    # multiply by scale factor
    mag *= c.scaling

    # apply elliptical corrections
    mat = Matrix3(
        Vector3(c.diag.x,    c.offdiag.x,  c.offdiag.y),
        Vector3(c.offdiag.x, c.diag.y,     c.offdiag.z),
        Vector3(c.offdiag.y, c.offdiag.z,  c.diag.z))

    mag = mat * mag

    # apply compassmot corrections
    if BAT is not None and hasattr(BAT, 'Curr') and not math.isnan(BAT.Curr):
        mag += c.cmot * BAT.Curr

    return mag

def get_yaw(ATT,MAG,BAT,c):
    '''calculate heading from raw magnetometer and new offsets'''

    mag = correct(MAG, BAT, c)

    # go via a DCM matrix to match the APM calculation
    dcm_matrix = mavextra.rotation_df(ATT)
    cos_pitch_sq = 1.0-(dcm_matrix.c.x*dcm_matrix.c.x)
    headY = mag.y * dcm_matrix.c.z - mag.z * dcm_matrix.c.y
    headX = mag.x * cos_pitch_sq - dcm_matrix.c.x * (mag.y * dcm_matrix.c.y + mag.z * dcm_matrix.c.z)

    global declination
    yaw = math.degrees(math.atan2(-headY,headX)) + declination
    if yaw < 0:
        yaw += 360
    return yaw

def expected_field(ATT, yaw):
    '''return expected magnetic field for attitude'''
    global earth_field

    roll = ATT.Roll
    pitch = ATT.Pitch

    rot = Matrix3()
    rot.from_euler(math.radians(roll), math.radians(pitch), math.radians(yaw))

    field = rot.transposed() * earth_field

    return field

data = None
old_corrections = Correction()

def wmm_error(p):
    '''world magnetic model error with correction fit'''
    p = list(p)
    c = copy.copy(old_corrections)

    c.offsets = Vector3(p.pop(0), p.pop(0), p.pop(0))
    c.scaling = p.pop(0)
    if margs['Elliptical']:
        c.diag = Vector3(p.pop(0), p.pop(0), p.pop(0))
        c.offdiag = Vector3(p.pop(0), p.pop(0), p.pop(0))
    else:
        c.diag = Vector3(1.0, 1.0, 1.0)
        c.offdiag = Vector3(0.0, 0.0, 0.0)

    if margs['CMOT']:
        c.cmot = Vector3(p.pop(0), p.pop(0), p.pop(0))

    ret = 0

    for (MAG,ATT,BAT) in data:
        yaw = get_yaw(ATT,MAG,BAT,c)
        expected = expected_field(ATT, yaw)
        observed = correct(MAG,BAT,c)

        error = (expected - observed).length()
        ret += error

    ret /= len(data)

    return ret

def fit_WWW():
    from scipy import optimize

    c = copy.copy(old_corrections)
    p = [c.offsets.x, c.offsets.y, c.offsets.z, c.scaling]
    if margs['Elliptical']:
        p.extend([c.diag.x, c.diag.y, c.diag.z, c.offdiag.x, c.offdiag.y, c.offdiag.z])
    if margs['CMOT']:
        p.extend([c.cmot.x, c.cmot.y, c.cmot.z])

    ofs = margs['OffsetMax']
    min_scale = margs['ScaleMin']
    max_scale = margs['ScaleMax']
    min_scale_delta = 0.00001
    bounds = [(-ofs,ofs),(-ofs,ofs),(-ofs,ofs),(min_scale,max(min_scale+min_scale_delta,max_scale))]
    if margs['CMOT NoChange']:
        bounds[0] = (c.offsets.x, c.offsets.x)
        bounds[1] = (c.offsets.y, c.offsets.y)
        bounds[2] = (c.offsets.z, c.offsets.z)

    if margs['Elliptical']:
        min_diag = margs['DiagonalMin']
        max_diag = margs['DiagonalMax']
        min_offdiag = margs['OffDiagMin']
        max_offdiag = margs['OffDiagMax']
        for i in range(3):
            bounds.append((min_diag,max_diag))
        for i in range(3):
            bounds.append((min_offdiag,max_offdiag))

    if margs['CMOT']:
        if margs['CMOT NoChange']:
            bounds.append((c.cmot.x, c.cmot.x))
            bounds.append((c.cmot.y, c.cmot.y))
            bounds.append((c.cmot.z, c.cmot.z))
        else:
            max_cmot = margs['CMOT Max']
            for i in range(3):
                bounds.append((-max_cmot,max_cmot))

    (p,err,iterations,imode,smode) = optimize.fmin_slsqp(wmm_error, p, bounds=bounds, full_output=True)
    if imode != 0:
        print("Fit failed: %s" % smode)
        sys.exit(1)
    p = list(p)

    c.offsets = Vector3(p.pop(0), p.pop(0), p.pop(0))
    c.scaling = p.pop(0)

    if margs['Elliptical']:
        c.diag = Vector3(p.pop(0), p.pop(0), p.pop(0))
        c.offdiag = Vector3(p.pop(0), p.pop(0), p.pop(0))
    else:
        c.diag = Vector3(1.0, 1.0, 1.0)
        c.offdiag = Vector3(0.0, 0.0, 0.0)

    if margs['CMOT']:
        c.cmot = Vector3(p.pop(0), p.pop(0), p.pop(0))
    else:
        c.cmot = Vector3(0.0, 0.0, 0.0)
    return c

def remove_offsets(MAG, BAT, c):
    '''remove all corrections to get raw sensor data'''
    correction_matrix = Matrix3(Vector3(c.diag.x,    c.offdiag.x, c.offdiag.y),
                                Vector3(c.offdiag.x, c.diag.y,    c.offdiag.z),
                                Vector3(c.offdiag.y, c.offdiag.z, c.diag.z))
    try:
        correction_matrix = correction_matrix.invert()
    except Exception:
        return False

    field = Vector3(MAG.MagX, MAG.MagY, MAG.MagZ)
    if BAT is not None and hasattr(BAT,'Curr') and not math.isnan(BAT.Curr):
        field -= c.cmot * BAT.Curr
    field = correction_matrix * field
    field *= 1.0 / c.scaling
    field -= Vector3(MAG.OfsX, MAG.OfsY, MAG.OfsZ)

    if math.isnan(field.x) or math.isnan(field.y) or math.isnan(field.z):
        return False
    MAG.MagX = int(field.x)
    MAG.MagY = int(field.y)
    MAG.MagZ = int(field.z)
    return True

def magfit(mlog, timestamp_in_range):
    '''find best magnetometer offset fit to a log file'''

    global earth_field, declination
    global data
    data = []

    ATT = None
    BAT = None

    mag_msg = margs['Magnetometer']
    global mag_idx
    if mag_msg[-1].isdigit():
        mag_instance = None
        mag_idx = mag_msg[-1]
    elif mag_msg.endswith('[0]'):
        mag_instance = 0
        mag_idx = ''
        mag_msg = 'MAG'
    elif mag_msg.endswith(']'):
        mag_instance = int(mag_msg[-2])
        mag_idx = str(mag_instance+1)
        mag_msg = 'MAG'
    else:
        mag_instance = None
        mag_idx = ''

    count = 0
    parameters = {}

    # get parameters
    mlog.rewind()
    while True:
        msg = mlog.recv_match(type=['PARM'])
        if msg is None:
            break
        parameters[msg.Name] = msg.Value

    mlog.rewind()

    lat = margs['Lattitude']
    lon = margs['Longitude']
    if lat != 0 and lon != 0:
        earth_field = mavextra.expected_earth_field_lat_lon(lat, lon)
        (declination,inclination,intensity) = mavextra.get_mag_field_ef(lat, lon)
        print("Earth field: %s  strength %.0f declination %.1f degrees" % (earth_field, earth_field.length(), declination))

    ATT_NAME = margs['Attitude']

    mtypes = ['GPS',mag_msg,ATT_NAME,'BAT']
    if ATT_NAME == "XKY0":
        mtypes.append('ATT')
    last_ATT = None
    print("Attitude source %s mtypes=%s" % (ATT_NAME, mtypes))

    # extract MAG data
    while True:
        msg = mlog.recv_match(type=mtypes)
        if msg is None:
            break
        in_range = timestamp_in_range(msg._timestamp)
        if in_range < 0:
            continue
        if in_range > 0:
            break
        if msg.get_type() == 'GPS' and msg.Status >= 3 and earth_field is None:
            earth_field = mavextra.expected_earth_field(msg)
            (declination,inclination,intensity) = mavextra.get_mag_field_ef(msg.Lat, msg.Lng)
            print("Earth field: %s  strength %.0f declination %.1f degrees" % (earth_field, earth_field.length(), declination))
        if msg.get_type() == 'ATT':
            # needed for XKY0 for yaw
            last_ATT = msg
        if msg.get_type() == ATT_NAME:
            if getattr(msg, 'C', 0) != 0:
                # use core zero for EKF attitude
                continue
            if ATT_NAME == 'XKY0':
                if last_ATT is None:
                    continue
                # get yaw from GSF, and roll/pitch from ATT
                ATT = last_ATT
                ATT.Yaw = math.degrees(msg.YC)
            else:
                ATT = msg
            ATT.Roll  += math.degrees(parameters['AHRS_TRIM_X'])
            ATT.Pitch += math.degrees(parameters['AHRS_TRIM_Y'])
            ATT.Yaw   += math.degrees(parameters['AHRS_TRIM_Z'])
        if msg.get_type() == 'BAT':
            if hasattr(msg,'Instance'):
                if margs['BatteryNum'] != msg.Instance+1:
                    continue
            BAT = msg
        if msg.get_type() == mag_msg and ATT is not None:
            if mag_instance is not None:
                if getattr(msg,'I',0) != mag_instance:
                    continue
            if count % margs['Reduce'] == 0:
                data.append((msg,ATT,BAT))
            count += 1

    old_corrections.offsets = Vector3(parameters.get('COMPASS_OFS%s_X' % mag_idx,0.0),
                                      parameters.get('COMPASS_OFS%s_Y' % mag_idx,0.0),
                                      parameters.get('COMPASS_OFS%s_Z' % mag_idx,0.0))
    old_corrections.diag = Vector3(parameters.get('COMPASS_DIA%s_X' % mag_idx,1.0),
                                   parameters.get('COMPASS_DIA%s_Y' % mag_idx,1.0),
                                   parameters.get('COMPASS_DIA%s_Z' % mag_idx,1.0))
    if old_corrections.diag == Vector3(0,0,0):
        old_corrections.diag = Vector3(1,1,1)
    old_corrections.offdiag = Vector3(parameters.get('COMPASS_ODI%s_X' % mag_idx,0.0),
                                      parameters.get('COMPASS_ODI%s_Y' % mag_idx,0.0),
                                      parameters.get('COMPASS_ODI%s_Z' % mag_idx,0.0))
    if parameters.get('COMPASS_MOTCT',0) == 2:
        # only support current based corrections for now
        old_corrections.cmot = Vector3(parameters.get('COMPASS_MOT%s_X' % mag_idx,0.0),
                                       parameters.get('COMPASS_MOT%s_Y' % mag_idx,0.0),
                                       parameters.get('COMPASS_MOT%s_Z' % mag_idx,0.0))
    old_corrections.scaling = parameters.get('COMPASS_SCALE%s' % mag_idx, None)
    if old_corrections.scaling is None or old_corrections.scaling < 0.1:
        force_scale = False
        old_corrections.scaling = 1.0
    else:
        force_scale = True

    # remove existing corrections
    data2 = []
    for (MAG,ATT,BAT) in data:
        if remove_offsets(MAG, BAT, old_corrections):
            data2.append((MAG,ATT,BAT))
    data = data2

    print("Extracted %u points" % len(data))
    print("Current: %s diag: %s offdiag: %s cmot: %s scale: %.2f" % (
        old_corrections.offsets, old_corrections.diag, old_corrections.offdiag, old_corrections.cmot, old_corrections.scaling))
    if len(data) == 0:
        return

    # do fit
    c = fit_WWW()

    # normalise diagonals to scale factor
    if force_scale:
        avgdiag = (c.diag.x + c.diag.y + c.diag.z)/3.0
        calc_scale = c.scaling
        c.scaling *= avgdiag
        min_scale = margs['ScaleMin']
        max_scale = margs['ScaleMax']
        if c.scaling > max_scale:
            c.scaling = max_scale
        if c.scaling < min_scale:
            c.scaling = min_scale
        scale_change = c.scaling / calc_scale
        c.diag *= 1.0/scale_change
        c.offdiag *= 1.0/scale_change

    print("New: %s diag: %s offdiag: %s cmot: %s scale: %.2f" % (
        c.offsets, c.diag, c.offdiag, c.cmot, c.scaling))

    x = []

    corrected = {}
    corrected['Yaw'] = []
    expected1 = {}
    expected2 = {}
    uncorrected = {}
    uncorrected['Yaw'] = []
    yaw_change1 = []
    yaw_change2 = []
    for i in range(len(data)):
        (MAG,ATT,BAT) = data[i]
        yaw1 = get_yaw(ATT,MAG,BAT,c)
        corrected['Yaw'].append(yaw1)
        ef1 = expected_field(ATT, yaw1)
        cf = correct(MAG, BAT, c)

        yaw2 = get_yaw(ATT,MAG,BAT,old_corrections)
        ef2 = expected_field(ATT, yaw2)
        uncorrected['Yaw'].append(yaw2)
        uf = correct(MAG, BAT, old_corrections)

        yaw_change1.append(mavextra.wrap_180(yaw1 - yaw2))
        yaw_change2.append(mavextra.wrap_180(yaw1 - ATT.Yaw))
        for axis in ['x','y','z']:
            if not axis in corrected:
                corrected[axis] = []
                uncorrected[axis] = []
                expected1[axis] = []
                expected2[axis] = []
            corrected[axis].append(getattr(cf, axis))
            uncorrected[axis].append(getattr(uf, axis))
            expected1[axis].append(getattr(ef1, axis))
            expected2[axis].append(getattr(ef2, axis))
        x.append(i)

    c.show_parms()

    fig, axs = pyplot.subplots(3, 1, sharex=True)

    for axis in ['x','y','z']:
        axs[0].plot(numpy.array(x), numpy.array(uncorrected[axis]), label='Uncorrected %s' % axis.upper() )
        axs[0].plot(numpy.array(x), numpy.array(expected2[axis]), label='Expected %s' % axis.upper() )
        axs[0].legend(loc='upper left')
        axs[0].set_title('Original')
        axs[0].set_ylabel('Field (mGauss)')

        axs[1].plot(numpy.array(x), numpy.array(corrected[axis]), label='Corrected %s' % axis.upper() )
        axs[1].plot(numpy.array(x), numpy.array(expected1[axis]), label='Expected %s' % axis.upper() )
        axs[1].legend(loc='upper left')
        axs[1].set_title('Corrected')
        axs[1].set_ylabel('Field (mGauss)')

    # show change in yaw estimate from old corrections to new
    axs[2].plot(numpy.array(x), numpy.array(yaw_change1), label='Mag Yaw Change')
    axs[2].plot(numpy.array(x), numpy.array(yaw_change2), label='ATT Yaw Change')
    axs[2].set_title('Yaw Change (degrees)')
    axs[2].legend(loc='upper left')

    pyplot.show(block=False)

class MagFit(MPDataLogChildTask):
    '''A class used to launch the MagFitUI in a child process'''

    def __init__(self, *args, **kwargs):
        '''
        Parameters
        ----------
        title : str
            The title of the application
        mlog : DFReader
            A dataflash or telemetry log
        xlimits: MAVExplorer.XLimits
            An object capturing timestamp limits
        '''

        super(MagFit, self).__init__(*args, **kwargs)

        # all attributes are implicitly passed to the child process 
        self.title = kwargs['title']
        self.xlimits = kwargs['xlimits']

    # @override
    def child_task(self):
        '''Launch the MagFitUI'''

        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx

        # create wx application
        app = wx.App(False)
        app.frame = MagFitUI(title=self.title,
                             close_event=self.close_event,
                             mlog=self.mlog,
                             timestamp_in_range=self.xlimits.timestamp_in_range)

        app.frame.SetDoubleBuffered(True)
        app.frame.Show()
        app.MainLoop()

class MagFitUI(wx.Dialog):
    def __init__(self, title, close_event, mlog, timestamp_in_range):
        super(MagFitUI, self).__init__(None, title=title, size=(600, 800))

        # capture the close event, log and timestamp range function
        self.close_event = close_event
        self.mlog = mlog
        self.timestamp_in_range = timestamp_in_range

        # events
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer, self.timer)
        self.timer.Start(100)
        self.Bind(wx.EVT_IDLE, self.OnIdle)

        # initialise the panels etc.
        self.init_ui()

    def OnIdle(self, event):
        time.sleep(0.05)

    def OnTimer(self, event):
        '''Periodically check if the close event has been received'''

        if self.close_event.wait(0.001):
            self.timer.Stop()
            self.Destroy()
            return

    def have_msg(self, msg):
        '''see if we have a given message name in the log'''
        mid = self.mlog.name_to_id.get(msg,-1)
        if mid == -1:
            return False
        return self.mlog.counts[mid] > 0

    def init_ui(self):
        '''Initalise the UI elements'''

        if not hasattr(self.mlog, 'formats'):
            print("Must be DF log")
            return

        self.panel = wx.Panel(self)
        self.vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox.AddStretchSpacer()
        self.panel.SetSizer(self.vbox)
        self.idmap = {}
        self.values = {}
        self.controls = {}
        self.callbacks = {}
        self.row = None

        msg_names = self.mlog.name_to_id.keys()
        mag_format = self.mlog.formats[self.mlog.name_to_id['MAG']]
        if 'I' in mag_format.columns:
            mag_choices = ['MAG[0]', 'MAG[1]', 'MAG[2]']
        else:
            mag_choices = ['MAG', 'MAG2', 'MAG3']

        att_choices = ['ATT']
        if self.have_msg('NKF1'):
            att_choices.append('NKF1')
        if self.have_msg('XKF1'):
            att_choices.append('XKF1')
        if self.have_msg('XKY0'):
            att_choices.append('XKY0')

        # first row, Mag and attitude source
        self.StartRow('Source Selection')
        self.AddCombo('Magnetometer', mag_choices)
        self.AddCombo('Attitude', att_choices)

        self.StartRow('Position')
        self.AddSpinFloat("Lattitude", -90, 90, 0.000001, 0, digits=8)
        self.AddSpinFloat("Longitude", -180, 180, 0.000001, 0, digits=8)

        self.StartRow()
        self.AddSpinInteger("Reduce", 1, 20, 1)

        self.StartRow('Offset Estimation')
        self.AddCheckBox("Offsets", default=True)
        self.StartRow()
        self.AddSpinInteger("OffsetMax", 500, 3000, 1500)

        self.StartRow('Scale Factor Estimation')
        self.AddSpinFloat("ScaleMin", 0.5, 2.0, 0.01, 1.0)
        self.AddSpinFloat("ScaleMax", 0.5, 2.0, 0.01, 1.0)

        self.StartRow('Elliptical Estimation')
        self.AddCheckBox("Elliptical")
        self.StartRow()
        self.AddSpinFloat("DiagonalMin", 0.8, 1.0, 0.01, 0.8)
        self.AddSpinFloat("DiagonalMax", 1.0, 1.2, 0.01, 1.2)
        self.StartRow()
        self.AddSpinFloat("OffDiagMin", -0.2, 0.0, 0.01, -0.2)
        self.AddSpinFloat("OffDiagMax", 0, 0.2, 0.01, 0.2)

        self.StartRow('Motor Interference Estimation')
        self.AddCheckBox("CMOT")
        self.AddCheckBox("CMOT NoChange")
        self.StartRow()
        self.AddSpinInteger("BatteryNum", 1, 8, 1)
        self.AddSpinFloat("CMOT Max", 1.0, 100, 0.1, 10)

        self.StartRow("Processing")
        self.AddButton('Run', callback=self.run)
        self.AddButton('Close', callback=self.close)
        self.EndRow()
        
        self.Center()

    def close(self):
        '''Set the close event'''

        self.close_event.set()

    def StartRow(self, label=None):
        if self.row:
            self.EndRow()
        if label:
            self.row = wx.BoxSizer(wx.HORIZONTAL)
            text = wx.StaticText(self.panel, label=label)
            font = wx.Font(16, wx.DEFAULT, wx.NORMAL, wx.BOLD)
            text.SetFont(font)
            self.row.Add(text, 0, wx.LEFT, 10)
            self.vbox.Add(self.row, 0, wx.TOP, 10)
        self.row = wx.BoxSizer(wx.HORIZONTAL)

    def EndRow(self):
        self.vbox.Add(self.row, 0, wx.TOP, 10)
        self.row = None

    def AddControl(self, c, label, default):
        self.idmap[c.GetId()] = label
        self.controls[c.GetId()] = c
        self.values[label] = default

    def AddCombo(self, label, choices):
        c = wx.ComboBox(choices=choices,
                        parent=self.panel,
                        style=0,
                        value=choices[0])
        self.AddControl(c, label, choices[0])
        c.Bind(wx.EVT_COMBOBOX, self.OnValue)
        self.row.Add(wx.StaticText(self.panel, label=label), 0, wx.LEFT, 20)
        self.row.Add(c, 0, wx.LEFT, 20)

    def AddButton(self, label, callback=None):
        c = wx.Button(self.panel, label=label)
        self.AddControl(c, label, False)
        if callback is not None:
            self.callbacks[c.GetId()] = callback
        c.Bind(wx.EVT_BUTTON, self.OnButton)
        self.row.Add(c, 0, wx.LEFT, 20)

    def AddCheckBox(self, label, default=False):
        c = wx.CheckBox(self.panel, label=label)
        c.SetValue(default)
        self.AddControl(c, label, default)
        c.Bind(wx.EVT_CHECKBOX, self.OnValue)
        self.row.Add(c, 0, wx.LEFT, 20)
        
    def AddSpinInteger(self, label, min_value, max_value, default):
        c = wx.SpinCtrl(self.panel, -1)
        c.SetRange(min_value, max_value)
        c.SetValue(default)
        self.AddControl(c, label, default)
        self.row.Add(wx.StaticText(self.panel, label=label), 0, wx.LEFT, 20)
        self.row.Add(c, 0, wx.LEFT, 20)
        self.row.Add(wx.StaticText(self.panel, label=""), 0, wx.LEFT, 20)
        self.Bind(wx.EVT_SPINCTRL, self.OnValue)

    def AddSpinFloat(self, label, min_value, max_value, increment, default, digits=2):
        c = wx.SpinCtrlDouble(self.panel, -1)
        c.SetRange(min_value, max_value)
        c.SetValue(default)
        c.SetIncrement(increment)
        c.SetDigits(digits)
        s1 = "%.*f" % (digits, min_value)
        s2 = "%.*f" % (digits, max_value)
        if len(s1) > len(s2):
            s = s1
        else:
            s = s2
        if hasattr(c, 'GetSizeFromText'):
            size = c.GetSizeFromText(s+" ")
            c.SetMinSize(size)
        self.AddControl(c, label, default)
        self.row.Add(wx.StaticText(self.panel, label=label), 0, wx.LEFT, 20)
        self.row.Add(c, 0, wx.LEFT, 20)
        self.row.Add(wx.StaticText(self.panel, label=""), 0, wx.LEFT, 20)
        self.Bind(wx.EVT_SPINCTRLDOUBLE, self.OnValue)
        
    def OnValue(self, event):
        self.values[self.idmap[event.GetId()]] = self.controls[event.GetId()].GetValue()

    def OnButton(self, event):
        self.values[self.idmap[event.GetId()]] = True
        if event.GetId() in self.callbacks:
            self.callbacks[event.GetId()]()

    def run(self):
        global margs
        margs = self.values
        magfit(self.mlog, self.timestamp_in_range)



#!/usr/bin/env python
'''test flight for DCM noise'''

import time, math

def enum(**enums):
    return type('Enum', (), enums)

TestState = enum(INIT=1, FBWA=2, AUTO=3)

from MAVProxy.modules.lib import mp_module

class TestModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(TestModule, self).__init__(mpstate, "test", "test flight")
        self.state = TestState.INIT
        print("Module test loaded")

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if self.state == TestState.INIT:
            if self.status.flightmode == "MANUAL":
                self.mpstate.functions.process_stdin("switch 4")
                self.mpstate.functions.process_stdin("rc 2 1300")
                self.mpstate.functions.process_stdin("rc 3 2000")
                self.mpstate.functions.process_stdin("module load sensors")
                self.mpstate.functions.process_stdin("watch sensors")
                self.mpstate.functions.process_stdin("wp list")
                self.state = TestState.FBWA
        if self.state == TestState.FBWA:
            if self.status.altitude > 60:
                self.mpstate.functions.process_stdin("rc 2 1500")
                self.mpstate.functions.process_stdin("auto")
                self.state = TestState.AUTO

def init(mpstate):
    '''initialise module'''
    return TestModule(mpstate)



"""
    MAVProxy rally module
"""

from pymavlink import mavwp
from pymavlink import mavutil
import time, os, platform
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util

if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

class RallyModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(RallyModule, self).__init__(mpstate, "rally", "rally point control", public = True)
        self.rallyloader_by_sysid = {}
        self.add_command('rally', self.cmd_rally, "rally point control", ["<add|clear|land|list|move|remove|>",
                                    "<load|save> (FILENAME)"])
        self.have_list = False
        self.abort_alt = 50
        self.abort_first_send_time = 0
        self.abort_previous_send_time = 0
        self.abort_ack_received = True

        self.menu_added_console = False
        self.menu_added_map = False
        if mp_util.has_wxpython:
            self.menu = MPMenuSubMenu('Rally',
                                  items=[MPMenuItem('Clear', 'Clear', '# rally clear'),
                                         MPMenuItem('List', 'List', '# rally list'),
                                         MPMenuItem('Load', 'Load', '# rally load ',
                                                    handler=MPMenuCallFileDialog(flags=('open',),
                                                                                 title='Rally Load',
                                                                                 wildcard='RallyPoints(*.txt,*.rally,*.ral)|*.txt;*.rally;*.ral')),
                                         MPMenuItem('Save', 'Save', '# rally save ',
                                                    handler=MPMenuCallFileDialog(flags=('save', 'overwrite_prompt'),
                                                                                 title='Rally Save',
                                                                                 wildcard='RallyPoints(*.txt,*.rally,*.ral)|*.txt;*.rally;*.ral')),
                                         MPMenuItem('Add', 'Add', '# rally add ',
                                                    handler=MPMenuCallTextDialog(title='Rally Altitude (m)',
                                                                                 default=100))])

    @property
    def rallyloader(self):
        '''rally loader by system ID'''
        if not self.target_system in self.rallyloader_by_sysid:
            self.rallyloader_by_sysid[self.target_system] = mavwp.MAVRallyLoader(self.settings.target_system,
                                                                                 self.settings.target_component)
        return self.rallyloader_by_sysid[self.target_system]

    def last_change(self):
        '''return time of last changes made to rally points'''
        return self.rallyloader.last_change

    def rally_count(self):
        '''return number of waypoints'''
        return self.rallyloader.rally_count()

    def rally_point(self, i):
        '''return instance of mavutil.mavlink.MAVLink_rally_point_message'''
        return self.rallyloader.rally_point(i)

    def set_last_change(self, time):
        '''can be used to cause map redraws'''
        self.rallyloader.last_change = time

    def idle_task(self):
        '''called on idle'''
        if self.module('console') is not None:
            if not self.menu_added_console:
                self.menu_added_console = True
                self.module('console').add_menu(self.menu)
        else:
            self.menu_added_console = False

        if self.module('map') is not None:
            if not self.menu_added_map:
                self.menu_added_map = True
                self.module('map').add_menu(self.menu)
        else:
            self.menu_added_map = False

        '''handle abort command; it is critical that the AP to receive it'''
        if self.abort_ack_received is False:
            #only send abort every second (be insistent, but don't spam)
            if (time.time() - self.abort_previous_send_time > 1):
                self.master.mav.command_long_send(self.settings.target_system,
                    self.settings.target_component,
                    mavutil.mavlink.MAV_CMD_DO_GO_AROUND,
                    0, int(self.abort_alt), 0, 0, 0, 0, 0, 0,)
                self.abort_previous_send_time = time.time()

            #try to get an ACK from the plane:
            if self.abort_first_send_time == 0:
                self.abort_first_send_time = time.time()
            elif time.time() - self.abort_first_send_time > 10: #give up after 10 seconds
                print("Unable to send abort command!\n")
                self.abort_ack_received = True


    def cmd_rally_add(self, args):
        '''handle rally add'''
        if len(args) < 1:
            alt = self.settings.rallyalt
        else:
            alt = float(args[0])

        if len(args) < 2:
            break_alt = self.settings.rally_breakalt
        else:
            break_alt = float(args[1])

        if len(args) < 3:
            flag = self.settings.rally_flags
        else:
            flag = int(args[2])
            #currently only supporting autoland values:
            #True (nonzero) and False (zero)
            if (flag != 0):
                flag = 2

        if not self.have_list:
            print("Please list rally points first")
            return

        if (self.rallyloader.rally_count() > 4):
            print("Only 5 rally points possible per flight plan.")
            return

        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return

        land_hdg = 0.0

        self.rallyloader.create_and_append_rally_point(latlon[0] * 1e7, latlon[1] * 1e7, alt, break_alt, land_hdg, flag)
        self.send_rally_points()
        print("Added Rally point at %s %f %f, autoland: %s" % (str(latlon), alt, break_alt, bool(flag & 2)))

    def cmd_rally_alt(self, args):
        '''handle rally alt change'''
        if (len(args) < 2):
            print("Usage: rally alt RALLYNUM newAlt <newBreakAlt>")
            return
        if not self.have_list:
            print("Please list rally points first")
            return

        idx = int(args[0])
        if idx <= 0 or idx > self.rallyloader.rally_count():
            print("Invalid rally point number %u" % idx)
            return

        new_alt = int(args[1])
        new_break_alt = None
        if (len(args) > 2):
            new_break_alt = int(args[2])

        self.rallyloader.set_alt(idx, new_alt, new_break_alt)
        self.send_rally_point(idx-1)
        self.fetch_rally_point(idx-1)
        self.rallyloader.reindex()

    def cmd_rally_move(self, args):
        '''handle rally move'''
        if len(args) < 1:
            print("Usage: rally move RALLYNUM")
            return
        if not self.have_list:
            print("Please list rally points first")
            return

        idx = int(args[0])
        if idx <= 0 or idx > self.rallyloader.rally_count():
            print("Invalid rally point number %u" % idx)
            return

        rpoint = self.rallyloader.rally_point(idx-1)

        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return

        oldpos = (rpoint.lat*1e-7, rpoint.lng*1e-7)
        self.rallyloader.move(idx, latlon[0], latlon[1])
        self.send_rally_point(idx-1)
        p = self.fetch_rally_point(idx-1)
        if p.lat != int(latlon[0]*1e7) or p.lng != int(latlon[1]*1e7):
            print("Rally move failed")
            return
        self.rallyloader.reindex()
        print("Moved rally point from %s to %s at %fm" % (str(oldpos), str(latlon), rpoint.alt))


    def cmd_rally(self, args):
        '''rally point commands'''
        #TODO: add_land arg
        if len(args) < 1:
            self.print_usage()
            return

        elif args[0] == "add":
            self.cmd_rally_add(args[1:])

        elif args[0] == "move":
            self.cmd_rally_move(args[1:])

        elif args[0] == "clear":
            self.rallyloader.clear()
            self.mav_param.mavset(self.master,'RALLY_TOTAL',0,3)

        elif args[0] == "remove":
            if not self.have_list:
                print("Please list rally points first")
                return
            if (len(args) < 2):
                print("Usage: rally remove RALLYNUM")
                return
            self.rallyloader.remove(int(args[1]))
            self.send_rally_points()

        elif args[0] == "list":
            self.list_rally_points()
            self.have_list = True

        elif args[0] == "load":
            if (len(args) < 2):
                print("Usage: rally load filename")
                return

            try:
                self.rallyloader.load(args[1].strip('"'))
            except Exception as msg:
                print("Unable to load %s - %s" % (args[1], msg))
                return

            self.send_rally_points()
            self.have_list = True

            print("Loaded %u rally points from %s" % (self.rallyloader.rally_count(), args[1]))

        elif args[0] == "save":
            if (len(args) < 2):
                print("Usage: rally save filename")
                return

            self.rallyloader.save(args[1].strip('"'))

            print("Saved rally file %s" % args[1])

        elif args[0] == "alt":
            self.cmd_rally_alt(args[1:])

        elif args[0] == "land":
            if (len(args) >= 2 and args[1] == "abort"):
                self.abort_ack_received = False
                self.abort_first_send_time = 0

                self.abort_alt = self.settings.rally_breakalt
                if (len(args) >= 3):
                    self.abort_alt = int(args[2])

            else:
                self.master.mav.command_long_send(self.settings.target_system,
                        self.settings.target_component,
                        mavutil.mavlink.MAV_CMD_DO_RALLY_LAND,
                        0, 0, 0, 0, 0, 0, 0, 0)

        else:
            self.print_usage()

    def mavlink_packet(self, m):
        '''handle incoming mavlink packet'''
        type = m.get_type()
        if type in ['COMMAND_ACK']:
            if m.command == mavutil.mavlink.MAV_CMD_DO_GO_AROUND:
                if (m.result == 0 and self.abort_ack_received == False):
                    self.say("Landing Abort Command Successfully Sent.")
                    self.abort_ack_received = True
                elif (m.result != 0 and self.abort_ack_received == False):
                    self.say("Landing Abort Command Unsuccessful.")

            elif m.command == mavutil.mavlink.MAV_CMD_DO_RALLY_LAND:
                if (m.result == 0):
                    self.say("Landing.")

    def unload(self):
        self.remove_command("rally")
        if self.module('console') is not None and self.menu_added_console:
            self.menu_added_console = False
            self.module('console').remove_menu(self.menu)
        if self.module('map') is not None and self.menu_added_map:
            self.menu_added_map = False
            self.module('map').remove_menu(self.menu)
        super(RallyModule, self).unload()

    def send_rally_point(self, i):
        '''send rally points from fenceloader'''
        p = self.rallyloader.rally_point(i)
        p.target_system = self.target_system
        p.target_component = self.target_component
        self.master.mav.send(p)

    def send_rally_points(self):
        '''send rally points from rallyloader'''
        self.mav_param.mavset(self.master,'RALLY_TOTAL',self.rallyloader.rally_count(),3)

        for i in range(self.rallyloader.rally_count()):
            self.send_rally_point(i)

    def fetch_rally_point(self, i):
        '''fetch one rally point'''
        self.master.mav.rally_fetch_point_send(self.target_system,
                                                    self.target_component, i)
        tstart = time.time()
        p = None
        while time.time() - tstart < 1:
            p = self.master.recv_match(type='RALLY_POINT', blocking=False)
            if p is not None:
                break
            time.sleep(0.1)
            continue
        if p is None:
            self.console.error("Failed to fetch rally point %u" % i)
            return None
        return p

    def list_rally_points(self):
        self.rallyloader.clear()
        rally_count = self.mav_param.get('RALLY_TOTAL',0)
        if rally_count == 0:
            print("No rally points")
            return
        for i in range(int(rally_count)):
            p = self.fetch_rally_point(i)
            if p is None:
                return
            self.rallyloader.append_rally_point(p)

        for i in range(self.rallyloader.rally_count()):
            p = self.rallyloader.rally_point(i)
            self.console.writeln("lat=%f lng=%f alt=%f break_alt=%f land_dir=%f autoland=%f" % (p.lat * 1e-7, p.lng * 1e-7, p.alt, p.break_alt, p.land_dir, int(p.flags & 2!=0) ))

        if self.logdir is not None:
            fname = 'ral.txt'
            if self.target_system > 1:
                fname = 'ral_%u.txt' % self.target_system
            ral_file_path = os.path.join(self.logdir, fname)
            self.rallyloader.save(ral_file_path)
            print("Saved rally points to %s" % ral_file_path)

    def print_usage(self):
        print("Usage: rally <list|load|land|save|add|remove|move|clear|alt>")

def init(mpstate):
    '''initialise module'''
    return RallyModule(mpstate)



#!/usr/bin/env python
'''
Arbitrary Message Module
Peter Barker, September 2017

'''

import time

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings


class message(mp_module.MPModule):
    def __init__(self, mpstate):
        """Initialise module"""
        super(message, self).__init__(mpstate, "message", "")
        self.status_callcount = 0
        self.boredom_interval = 10  # seconds
        self.last_bored = time.time()

        self.packets_mytarget = 0
        self.packets_othertarget = 0
        self.verbose = False

        self.message_settings = mp_settings.MPSettings(
            [('verbose', bool, False)])
        self.add_command('message', self.cmd_message, "message module", [])

    def usage(self):
        '''show help on command line options'''
        return "Usage: message TYPE ARG..."

    def cmd_message(self, args):
        if len(args) == 0:
            print(self.usage())
        else:
            packettype = args[0]
            methodname = packettype.lower() + "_send"
            transformed = [eval(x) for x in args[1:]]
            try:
                method = getattr(self.master.mav, methodname)
            except AttributeError:
                print("Unable to find %s" % methodname)
                return
            method(*transformed)


def init(mpstate):
    '''initialise module'''
    return message(mpstate)


# STABILIZE> message MISSION_CLEAR_ALL int(1) int(1)
# STABILIZE> Got MAVLink msg: MISSION_ACK {target_system : 255,
# target_component : 0, type : 0}



'''
Support for AIS data
Stephen Dade
June 2021
'''

from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import mp_util

if mp_util.has_wxpython:
    from MAVProxy.modules.lib import mp_menu
    from MAVProxy.modules.mavproxy_map import mp_slipmap

class AISVehicle():
    '''a generic AIS threat'''

    def __init__(self, id, state):
        self.id = id
        self.state = state
        self.vehicle_colour = 'green'  # use boat icon for now
        self.vehicle_type = 'boat'
        self.icon = self.vehicle_colour + self.vehicle_type + '.png'
        self.update_time = 0
        self.onMap = False

    def update(self, state, tnow):
        self.state = state
        self.update_time = tnow

    def getThreatRadius(self, default_radius):
        ''' get threat radius, based on ship type'''
        # threat radius is the maximum of distance to AIS location
        if self.state.flags & mavutil.mavlink.AIS_FLAGS_LARGE_BOW_DIMENSION or self.state.flags & mavutil.mavlink.AIS_FLAGS_LARGE_STERN_DIMENSION:
            threat_radius = 511
        elif self.state.flags & mavutil.mavlink.AIS_FLAGS_LARGE_PORT_DIMENSION or self.state.flags & mavutil.mavlink.AIS_FLAGS_LARGE_STARBOARD_DIMENSION:
            threat_radius = 63
        else:
            threat_radius = max(self.state.dimension_bow, self.state.dimension_stern, self.state.dimension_port, self.state.dimension_starboard)
        #default to threat_radius setting
        if threat_radius == 0:
            return default_radius

class AISModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(AISModule, self).__init__(mpstate, "ais", "AIS data support", public=True)
        self.threat_vehicles = {}

        self.add_command('ais', self.cmd_AIS, "ais control",
                         ["<status>", "set (AISSETTING)"])

        self.AIS_settings = mp_settings.MPSettings([("timeout", int, 60),  # seconds
                                                    ("threat_radius", int, 200)])  # meters
        self.add_completion_function('(AISSETTING)',
                                     self.AIS_settings.completion)

        self.threat_timeout_timer = mavutil.periodic_event(2)

        self.update_map_timer = mavutil.periodic_event(1)

        self.tnow = self.get_time()

    def cmd_AIS(self, args):
        '''ais command parser'''
        usage = "usage: ais <set>"
        if len(args) == 0:
            print(usage)
        elif args[0] == "set":
            self.AIS_settings.command(args[1:])
        else:
            print(usage)

    def check_threat_timeout(self):
        '''check and handle threat time out'''
        for id in self.threat_vehicles:
            if self.threat_vehicles[id].update_time == 0:
                self.threat_vehicles[id].update_time = self.get_time()
            dt = self.get_time() - self.threat_vehicles[id].update_time
            if dt > self.AIS_settings.timeout:
                # if the threat has timed out...
                del self.threat_vehicles[id]  # remove the threat from the dict
                for mp in self.module_matching('map*'):
                    # remove the threat from the map
                    mp.map.remove_object(id)
                    mp.map.remove_object(id+":circle")
                # we've modified the dict we're iterating over, so
                # we'll get any more timed-out threats next time we're
                # called:
                return

    def update_map(self):
        '''update the map graphics'''
        for mp in self.module_matching('map*'):
            for id in self.threat_vehicles.keys():
                mstate = self.threat_vehicles[id].state
                threat_radius = self.threat_vehicles[id].getThreatRadius(self.AIS_settings.threat_radius)
                # update if existing object on map, else create
                if self.threat_vehicles[id].onMap:
                    mp.map.set_position(id, (mstate.lat * 1e-7, mstate.lon * 1e-7), 3, rotation=mstate.heading*0.01, label=str(mstate.callsign))
                    # Turns out we can't edit the circle's radius, so have to remove and re-add
                    mp.map.remove_object(id+":circle")
                    mp.map.add_object(mp_slipmap.SlipCircle(id+":circle", 3,
                                                            (mstate.lat * 1e-7, mstate.lon * 1e-7),
                                                            threat_radius, (0, 255, 255), linewidth=1))
                else:
                    self.threat_vehicles[id].menu_item = mp_menu.MPMenuItem(name=id, returnkey=None)

                    # draw the vehicle on the map
                    popup = mp_menu.MPMenuSubMenu('AIS', items=[self.threat_vehicles[id].menu_item])
                    icon = mp_slipmap.SlipIcon(
                        id,
                        (mstate.lat * 1e-7, mstate.lon * 1e-7),
                        mp.map.icon(self.threat_vehicles[id].icon),
                        3,
                        rotation=mstate.heading*0.01,
                        follow=False,
                        trail=mp_slipmap.SlipTrail(colour=(0, 255, 255)),
                        popup_menu=popup,
                    )
                    mp.map.add_object(icon)
                    mp.map.add_object(mp_slipmap.SlipCircle(id+":circle", 3,
                                                            (mstate.lat * 1e-7, mstate.lon * 1e-7),
                                                            threat_radius, (0, 255, 255), linewidth=1))
                    self.threat_vehicles[id].onMap = True

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == "AIS_VESSEL":
            id = 'AIS-' + str(m.MMSI)

            if id not in self.threat_vehicles.keys():  # check to see if the vehicle is in the dict
                # if not then add it
                self.threat_vehicles[id] = AISVehicle(id=id, state=m)
            else:  # the vehicle is in the dict
                # update the dict entry
                self.threat_vehicles[id].update(m, self.get_time())

    def idle_task(self):
        '''called on idle'''
        if self.threat_timeout_timer.trigger():
            self.check_threat_timeout()
        if self.update_map_timer.trigger():
            self.update_map()

def init(mpstate):
    '''initialise module'''
    return AISModule(mpstate)



'''
support for asterix SDPS data, setup for OBC 2018

This listens for SDPS on UDP and translates to ADSB_VEHICLE messages
'''

import pickle
from math import *

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import mp_util
from pymavlink import mavutil

import asterix, socket, time, os, struct

class Track:
    def __init__(self, adsb_pkt):
        self.pkt = adsb_pkt
        self.last_time = 0

    def update(self, pkt, tnow):
        dt = tnow - self.last_time
        if dt < 0 or dt > 10:
            self.last_time = tnow
            return
        if dt < 0.1:
            return
        self.last_time = tnow
        dist = mp_util.gps_distance(self.pkt.lat*1e-7, self.pkt.lon*1e-7,
                                    pkt.lat*1e-7, pkt.lon*1e-7)
        if dist > 0.01:
            heading = mp_util.gps_bearing(self.pkt.lat*1e-7, self.pkt.lon*1e-7,
                                        pkt.lat*1e-7, pkt.lon*1e-7)
            spd = dist / dt
            pkt.heading = int(heading*100)
            new_vel = int(spd * 100)
            #print(pkt.ICAO_address, new_vel*0.01, pkt.hor_velocity*0.01)
            pkt.hor_velocity = new_vel
        if pkt.hor_velocity > 65535:
            pkt.hor_velocity = 65535
        self.pkt = pkt

class VehiclePos(object):
    def __init__(self, GPI):
        self.lat = GPI.lat * 1.0e-7
        self.lon = GPI.lon * 1.0e-7
        self.alt = GPI.alt * 1.0e-3
        self.vx = GPI.vx * 1.0e-2
        self.vy = GPI.vy * 1.0e-2

class AsterixModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(AsterixModule, self).__init__(mpstate, "asterix", "asterix SDPS data support")
        self.threat_vehicles = {}
        self.active_threat_ids = []  # holds all threat ids the vehicle is evading

        self.add_command('asterix', self.cmd_asterix, "asterix control",
                         ["<start|stop>","set (ASTERIXSETTING)"])

        # filter_dist is distance in metres
        self.asterix_settings = mp_settings.MPSettings([("port", int, 45454),
                                                        ('debug', int, 0),
                                                        ('filter_dist_xy', int, 1000),
                                                        ('filter_dist_z', int, 250),
                                                        ('filter_time', int, 20),
                                                        ('wgs84_to_AMSL', float, -41.2),
                                                        ('filter_use_vehicle2', bool, True),
        ])
        self.add_completion_function('(ASTERIXSETTING)',
                                     self.asterix_settings.completion)
        self.sock = None
        self.tracks = {}
        self.start_listener()

        # storage for vehicle positions, used for filtering
        self.vehicle_pos = None
        self.vehicle2_pos = None

        self.adsb_packets_sent = 0
        self.adsb_packets_not_sent = 0
        self.adsb_byterate = 0 # actually bytes...
        self.adsb_byterate_update_timestamp = 0
        self.adsb_last_packets_sent = 0
        if self.logdir is not None:
            logpath = os.path.join(self.logdir, 'asterix.log')
        else:
            logpath = 'asterix.log'
        self.logfile = open(logpath, 'wb')
        self.pkt_count = 0
        self.console.set_status('ASTX', 'ASTX --/--', row=6)

    def print_status(self):
        print("ADSB packets sent: %u" % self.adsb_packets_sent)
        print("ADSB packets not sent: %u" % self.adsb_packets_not_sent)
        print("ADSB bitrate: %u bytes/s" % int(self.adsb_byterate))

    def cmd_asterix(self, args):
        '''asterix command parser'''
        usage = "usage: asterix <set|start|stop|restart|status>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.asterix_settings.command(args[1:])
        elif args[0] == "start":
            self.start_listener()
        elif args[0] == "stop":
            self.stop_listener()
        elif args[0] == "restart":
            self.stop_listener()
            self.start_listener()
        elif args[0] == "status":
            self.print_status()
        else:
            print(usage)

    def start_listener(self):
        '''start listening for packets'''
        if self.sock is not None:
            self.sock.close()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(('', self.asterix_settings.port))
        self.sock.setblocking(False)
        print("Started on port %u" % self.asterix_settings.port)

    def stop_listener(self):
        '''stop listening for packets'''
        if self.sock is not None:
            self.sock.close()
            self.sock = None
        self.tracks = {}

    def set_secondary_vehicle_position(self, m):
        '''store second vehicle position for filtering purposes'''
        if m.get_type() != 'GLOBAL_POSITION_INT':
            return
        (lat, lon, heading) = (m.lat*1.0e-7, m.lon*1.0e-7, m.hdg*0.01)
        if abs(lat) < 1.0e-3 and abs(lon) < 1.0e-3:
            return
        self.vehicle2_pos = VehiclePos(m)

    def could_collide_hor(self, vpos, adsb_pkt):
        '''return true if vehicle could come within filter_dist_xy meters of adsb vehicle in timeout seconds'''
        margin = self.asterix_settings.filter_dist_xy
        timeout = self.asterix_settings.filter_time
        alat = adsb_pkt.lat * 1.0e-7
        alon = adsb_pkt.lon * 1.0e-7
        avel = adsb_pkt.hor_velocity * 0.01
        vvel = sqrt(vpos.vx**2 + vpos.vy**2)
        dist = mp_util.gps_distance(vpos.lat, vpos.lon, alat, alon)
        dist -= avel * timeout
        dist -= vvel * timeout
        if dist <= margin:
            return True
        return False

    def could_collide_ver(self, vpos, adsb_pkt):
        '''return true if vehicle could come within filter_dist_z meters of adsb vehicle in timeout seconds'''
        if adsb_pkt.emitter_type < 100 or adsb_pkt.emitter_type > 104:
            return True
        margin = self.asterix_settings.filter_dist_z
        vtype = adsb_pkt.emitter_type - 100
        valt = vpos.alt
        aalt1 = adsb_pkt.altitude * 0.001
        if vtype == 2:
            # weather, always yes
            return True
        if vtype == 4:
            # bird of prey, always true
            return True
        # planes and migrating birds have 150m margin
        aalt2 = aalt1 + adsb_pkt.ver_velocity * 0.01 * self.asterix_settings.filter_time
        altsep1 = abs(valt - aalt1)
        altsep2 = abs(valt - aalt2)
        if altsep1 > 150 + margin and altsep2 > 150 + margin:
            return False
        return True

    def should_send_adsb_pkt(self, adsb_pkt):
        if self.vehicle_pos is not None:
            if self.could_collide_hor(self.vehicle_pos, adsb_pkt) and self.could_collide_ver(self.vehicle_pos, adsb_pkt):
                return True

        if self.vehicle2_pos is not None:
            if self.could_collide_hor(self.vehicle2_pos, adsb_pkt) and self.could_collide_ver(self.vehicle2_pos, adsb_pkt):
                return True

        return False

    def idle_task(self):
        '''called on idle'''
        if self.sock is None:
            return
        try:
            pkt = self.sock.recv(10240)
        except Exception:
            return
        try:
            if pkt.startswith(b'PICKLED:'):
                pkt = pkt[8:]
                # pickled packet
                try:
                    amsg = [pickle.loads(pkt)]
                except pickle.UnpicklingError:
                    amsg = asterix.parse(pkt)
            else:
                amsg = asterix.parse(pkt)
            self.pkt_count += 1
            self.console.set_status('ASTX', 'ASTX %u/%u' % (self.pkt_count, self.adsb_packets_sent), row=6)
        except Exception:
            print("bad packet")
            return
        try:
            logpkt = b'AST:' + struct.pack('<dI', time.time(), len(pkt)) + pkt
            self.logfile.write(logpkt)
        except Exception:
            pass
        
        for m in amsg:
            if self.asterix_settings.debug > 1:
                print(m)
            lat = m['I105']['Lat']['val']
            lon = m['I105']['Lon']['val']
            alt_f = m['I130']['Alt']['val']
            climb_rate_fps = m['I220']['RoC']['val']
            sac = m['I010']['SAC']['val']
            sic = m['I010']['SIC']['val']
            trkn = m['I040']['TrkN']['val']
            # fake ICAO_address
            icao_address = trkn & 0xFFFF

            # use squawk for time in 0.1 second increments. This allows for old msgs to be discarded on vehicle
            # when using more than one link to vehicle
            squawk = (int(self.mpstate.attitude_time_s * 10) & 0xFFFF)

            alt_m = alt_f * 0.3048

            # asterix is WGS84, ArduPilot uses AMSL, which is EGM96
            alt_m += self.asterix_settings.wgs84_to_AMSL

            # consider filtering this packet out; if it's not close to
            # either home or the vehicle position don't send it
            adsb_pkt = self.master.mav.adsb_vehicle_encode(icao_address,
                                                           int(lat*1e7),
                                                           int(lon*1e7),
                                                           mavutil.mavlink.ADSB_ALTITUDE_TYPE_GEOMETRIC,
                                                           int(alt_m*1000), # mm
                                                           0, # heading
                                                           0, # hor vel
                                                           int(climb_rate_fps * 0.3048 * 100), # cm/s
                                                           ("%08x" % icao_address).encode("ascii"),
                                                           100 + (trkn // 10000),
                                                           1,
                                                           (mavutil.mavlink.ADSB_FLAGS_VALID_COORDS |
                                                            mavutil.mavlink.ADSB_FLAGS_VALID_ALTITUDE |
                                                            mavutil.mavlink.ADSB_FLAGS_VALID_VELOCITY |
                                                            mavutil.mavlink.ADSB_FLAGS_VALID_HEADING),
                                                           squawk)
            if icao_address in self.tracks:
                self.tracks[icao_address].update(adsb_pkt, self.get_time())
            else:
                self.tracks[icao_address] = Track(adsb_pkt)
            if self.asterix_settings.debug > 0:
                print(adsb_pkt)
            # send on all links
            if self.should_send_adsb_pkt(adsb_pkt):
                self.adsb_packets_sent += 1
                for i in range(len(self.mpstate.mav_master)):
                    conn = self.mpstate.mav_master[i]
                    #if adsb_pkt.hor_velocity < 1:
                    #    print(adsb_pkt)
                    conn.mav.send(adsb_pkt)
            else:
                self.adsb_packets_not_sent += 1

            adsb_mod = self.module('adsb')
            if adsb_mod:
                # the adsb module is loaded, display on the map
                adsb_mod.mavlink_packet(adsb_pkt)

            try:
                for sysid in self.mpstate.sysid_outputs:
                    # fwd to sysid clients
                    adsb_pkt.pack(self.mpstate.sysid_outputs[sysid].mav)
                    self.mpstate.sysid_outputs[sysid].write(adsb_pkt.get_msgbuf())
            except Exception:
                pass
                
        now = time.time()
        delta = now - self.adsb_byterate_update_timestamp
        if delta > 5:
            self.adsb_byterate_update_timestamp = now
            bytes_per_adsb_packet = 38 # FIXME: find constant
            self.adsb_byterate = (self.adsb_packets_sent - self.adsb_last_packets_sent)/delta * bytes_per_adsb_packet
            self.adsb_last_packets_sent = self.adsb_packets_sent

    def mavlink_packet(self, m):
        '''get time from mavlink ATTITUDE'''
        if m.get_type() == 'GLOBAL_POSITION_INT':
            if abs(m.lat) < 1000 and abs(m.lon) < 1000:
                return
            self.vehicle_pos = VehiclePos(m)

def init(mpstate):
    '''initialise module'''
    return AsterixModule(mpstate)

if __name__ == '__main__':
    import sys
    logname = sys.argv[1]
    logf = open(logname, 'rb')
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.connect(('', 45454))
    t0 = None
    tstart = time.time()
    while True:
        header = logf.read(16)
        if header[0:4] != 'AST:':
            print("Bad header", header[0:4])
            break
        (t,len) = struct.unpack('<dI', header[4:16])
        pkt = logf.read(len)
        if t0 is None:
            t0 = t
        tdiff = t - t0
        time.sleep(tdiff)
        t0 = t
        sock.send(pkt)
        #amsg = asterix.parse(pkt)
        #print(amsg)





#!/usr/bin/env python
'''
mission editor module
Michael Day
June 2104
'''

from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.lib import win_layout

from MAVProxy.modules.mavproxy_misseditor import me_event
MissionEditorEvent = me_event.MissionEditorEvent

from pymavlink import mavutil
from pymavlink import mavwp

import time
import threading

class MissionEditorEventThread(threading.Thread):
    def __init__(self, mp_misseditor, q, l):
        threading.Thread.__init__(self)
        self.mp_misseditor = mp_misseditor
        self.event_queue = q
        self.event_queue_lock = l
        self.time_to_quit = False

    def module(self, name):
        '''access another module'''
        return self.mp_misseditor.mpstate.module(name)

    def master(self):
        '''access master mavlink connection'''
        return self.mp_misseditor.mpstate.master()
    
    def run(self):
        while not self.time_to_quit:
            queue_access_start_time = time.time()
            self.event_queue_lock.acquire()
            request_read_after_processing_queue = False
            while (not self.event_queue.empty()) and (time.time() - queue_access_start_time) < 0.6:
                event = self.event_queue.get()

                if isinstance(event, win_layout.WinLayout):
                    win_layout.set_layout(event, self.mp_misseditor.set_layout)
                elif isinstance(event, mavwp.MAVWPLoader):
                    self.mp_misseditor.gui_event_queue_lock.acquire()
                    self.mp_misseditor.gui_event_queue.put(MissionEditorEvent(
                        me_event.MEGE_CLEAR_MISS_TABLE))
                    self.mp_misseditor.gui_event_queue_lock.release()
                    if event.count() > 0:
                        self.mp_misseditor.gui_event_queue_lock.acquire()
                        self.mp_misseditor.gui_event_queue.put(MissionEditorEvent(
                            me_event.MEGE_ADD_MISS_TABLE_ROWS,num_rows=event.count()-1))
                        self.mp_misseditor.gui_event_queue_lock.release()

                    for m in event.wpoints:
                        self.mp_misseditor.gui_event_queue_lock.acquire()
                        self.mp_misseditor.gui_event_queue.put(MissionEditorEvent(
                            me_event.MEGE_SET_MISS_ITEM,
                            num=m.seq,command=m.command,param1=m.param1,
                            param2=m.param2,param3=m.param3,param4=m.param4,
                            lat=m.x,lon=m.y,alt=m.z,frame=m.frame))
                        self.mp_misseditor.gui_event_queue_lock.release()
                else:
                    event_type = event.get_type()

                    if event_type == me_event.MEE_READ_WPS:
                        self.module('wp').cmd_wp(['list'])
                        #list the rally points while I'm add it:
                        #TODO: DON'T KNOW WHY THIS DOESN'T WORK
                        #self.module('rally').cmd_rally(['list'])

                        #means I'm doing a read & don't know how many wps to expect:
                        self.mp_misseditor.num_wps_expected = -1
                        self.wps_received = {}

                    elif event_type == me_event.MEE_TIME_TO_QUIT:
                        self.time_to_quit = True

                    elif event_type == me_event.MEE_GET_WP_RAD:
                        wp_radius = self.module('param').mav_param.get('WP_RADIUS')
                        if (wp_radius is None):
                            continue
                        self.mp_misseditor.gui_event_queue_lock.acquire()
                        self.mp_misseditor.gui_event_queue.put(MissionEditorEvent(
                            me_event.MEGE_SET_WP_RAD,wp_rad=wp_radius))
                        self.mp_misseditor.gui_event_queue_lock.release()

                    elif event_type == me_event.MEE_SET_WP_RAD:
                        self.mp_misseditor.param_set('WP_RADIUS',event.get_arg("rad"))

                    elif event_type == me_event.MEE_GET_LOIT_RAD:
                        loiter_radius = self.module('param').mav_param.get('WP_LOITER_RAD')
                        if (loiter_radius is None):
                            continue
                        self.mp_misseditor.gui_event_queue_lock.acquire()
                        self.mp_misseditor.gui_event_queue.put(MissionEditorEvent(
                            me_event.MEGE_SET_LOIT_RAD,loit_rad=loiter_radius))
                        self.mp_misseditor.gui_event_queue_lock.release()

                    elif event_type == me_event.MEE_SET_LOIT_RAD:
                        loit_rad = event.get_arg("rad")
                        if (loit_rad is None):
                            continue

                        self.mp_misseditor.param_set('WP_LOITER_RAD', loit_rad)

                        #need to redraw rally points
                        # Don't understand why this rally refresh isn't lagging...
                        # likely same reason why "timeout setting WP_LOITER_RAD"
                        #comes back:
                        #TODO: fix timeout issue
                        self.module('rally').set_last_change(time.time())

                    elif event_type == me_event.MEE_GET_WP_DEFAULT_ALT:
                        self.mp_misseditor.gui_event_queue_lock.acquire()
                        self.mp_misseditor.gui_event_queue.put(MissionEditorEvent(
                            me_event.MEGE_SET_WP_DEFAULT_ALT,def_wp_alt=self.mp_misseditor.mpstate.settings.wpalt))
                        self.mp_misseditor.gui_event_queue_lock.release()
                    elif event_type == me_event.MEE_SET_WP_DEFAULT_ALT:
                        self.mp_misseditor.mpstate.settings.command(["wpalt",event.get_arg("alt")])

                    elif event_type == me_event.MEE_WRITE_WPS:
                        self.module('wp').wploader.clear()
                        self.master().waypoint_count_send(event.get_arg("count"))
                        self.mp_misseditor.num_wps_expected = event.get_arg("count")
                        self.mp_misseditor.wps_received = {}
                    elif event_type == me_event.MEE_WRITE_WP_NUM:
                        w = mavutil.mavlink.MAVLink_mission_item_message(
                            self.mp_misseditor.mpstate.settings.target_system,
                            self.mp_misseditor.mpstate.settings.target_component,
                            event.get_arg("num"),
                            int(event.get_arg("frame")),
                            event.get_arg("cmd_id"),
                            0, 1,
                            event.get_arg("p1"), event.get_arg("p2"),
                            event.get_arg("p3"), event.get_arg("p4"),
                            event.get_arg("lat"), event.get_arg("lon"),
                            event.get_arg("alt"))

                        self.module('wp').wploader.add(w)
                        wsend = self.module('wp').wploader.wp(w.seq)
                        if self.mp_misseditor.mpstate.settings.wp_use_mission_int:
                            wsend = self.module('wp').wp_to_mission_item_int(w)
                        self.master().mav.send(wsend)

                        #tell the wp module to expect some waypoints
                        self.module('wp').loading_waypoints = True

                    elif event_type == me_event.MEE_LOAD_WP_FILE:
                        self.module('wp').cmd_wp(['load',event.get_arg("path")])
                        #Wait for the other thread to finish loading waypoints.
                        #don't let this loop run forever in case we have a lousy
                        #link to the plane
                        i = 0
                        while (i < 10 and
                               self.module('wp').loading_waypoints):
                            time.sleep(1)
                            i = i + 1

                            #don't modify queue while in the middile of processing it:
                            request_read_after_processing_queue = True

                    elif event_type == me_event.MEE_SAVE_WP_FILE:
                        self.module('wp').cmd_wp(['save',event.get_arg("path")])

            self.event_queue_lock.release()

            #if event processing operations require a mission referesh in GUI
            #(e.g., after a load or a verified-completed write):
            if (request_read_after_processing_queue):
                self.event_queue_lock.acquire()
                self.event_queue.put(MissionEditorEvent(me_event.MEE_READ_WPS))
                self.event_queue_lock.release()

            #periodically re-request WPs that were never received:
            #DON'T NEED TO! -- wp module already doing this

            time.sleep(0.2)

class MissionEditorMain(object):
    def __init__(self, mpstate):
        self.num_wps_expected = 0 #helps me to know if all my waypoints I'm expecting have arrived
        self.wps_received = {}

        self.event_queue = multiproc.Queue()
        self.event_queue_lock = multiproc.Lock()
        self.gui_event_queue = multiproc.Queue()
        self.gui_event_queue_lock = multiproc.Lock()

        self.object_queue = multiproc.Queue()

        self.close_window = multiproc.Semaphore()
        self.close_window.acquire()

        self.child = multiproc.Process(target=self.child_task,args=(self.event_queue,self.event_queue_lock,self.gui_event_queue,self.gui_event_queue_lock,self.close_window))
        self.child.start()

        self.event_thread = MissionEditorEventThread(self, self.event_queue, self.event_queue_lock)
        self.event_thread.start()

        self.mpstate = mpstate
        self.mpstate.miss_editor = self

        self.last_unload_check_time = time.time()
        self.unload_check_interval = 0.1 # seconds

        self.time_to_quit = False
        self.mavlink_message_queue = multiproc.Queue()
        self.mavlink_message_queue_handler = threading.Thread(target=self.mavlink_message_queue_handler)
        self.mavlink_message_queue_handler.start()
        self.needs_unloading = False
        self.last_wp_change = time.time()

    def mavlink_message_queue_handler(self):
        while not self.time_to_quit:
            while True:
                if self.time_to_quit:
                    return
                if not self.mavlink_message_queue.empty():
                    break
                time.sleep(0.1)
            m = self.mavlink_message_queue.get()

            #MAKE SURE YOU RELEASE THIS LOCK BEFORE LEAVING THIS METHOD!!!
            #No "return" statement should be put in this method!
            self.gui_event_queue_lock.acquire()

            try:
                self.process_mavlink_packet(m)
            except Exception as e:
                print("Caught exception (%s)" % str(e))
                import traceback
                traceback.print_stack()

            self.gui_event_queue_lock.release()

    def unload(self):
        '''unload module'''
        self.mpstate.miss_editor.close()
        self.mpstate.miss_editor = None

    def get_wps_from_module(self):
        '''get WP list from wp module'''
        self.event_queue_lock.acquire()
        self.event_queue.put(self.mpstate.module('wp').wploader)
        self.event_queue_lock.release()

    def idle_task(self):
        now = time.time()
        if self.last_unload_check_time + self.unload_check_interval < now:
            self.last_unload_check_time = now
            if not self.child.is_alive():
                self.close()
                return
        last_wp_change = self.mpstate.module('wp').loading_waypoint_lasttime
        if last_wp_change > self.last_wp_change:
            self.last_wp_change = last_wp_change
            self.get_wps_from_module()




    def mavlink_packet(self, m):
        if (getattr(m, 'mission_type', None) is not None and
            m.mission_type != mavutil.mavlink.MAV_MISSION_TYPE_MISSION):
            return
        mtype = m.get_type()
        if mtype in ['WAYPOINT_COUNT','MISSION_COUNT', 'WAYPOINT', 'MISSION_ITEM', 'MISSION_ITEM_INT']:
            if mtype == 'MISSION_ITEM_INT':
                m = self.mpstate.module('wp').wp_from_mission_item_int(m)
            self.mavlink_message_queue.put(m)

    def process_mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        mtype = m.get_type()

        # if you add processing for an mtype here, remember to add it
        # to mavlink_packet, above
        if (getattr(m, 'mission_type', None) is not None and
            m.mission_type != mavutil.mavlink.MAV_MISSION_TYPE_MISSION):
            return
        if mtype in ['WAYPOINT_COUNT','MISSION_COUNT']:
            if (self.num_wps_expected == 0):
                #I haven't asked for WPs, or these messages are duplicates
                #of msgs I've already received.
                self.mpstate.console.error("No waypoint load started (from Editor).")
            #I only clear the mission in the Editor if this was a read event
            elif (self.num_wps_expected == -1):
                self.gui_event_queue.put(MissionEditorEvent(
                    me_event.MEGE_CLEAR_MISS_TABLE))
                self.num_wps_expected = m.count
                self.wps_received = {}

                if m.count > 1:
                    self.gui_event_queue.put(MissionEditorEvent(
                        me_event.MEGE_ADD_MISS_TABLE_ROWS,num_rows=m.count-1))
            #write has been sent by the mission editor:
            elif (self.num_wps_expected > 1):
                if (m.count != self.num_wps_expected):
                    self.mpstate.console.error("Unexpected waypoint count from vehicle after write (Editor)")
                #since this is a write operation from the Editor there
                #should be no need to update number of table rows

        elif mtype in ['WAYPOINT', 'MISSION_ITEM']:
            #still expecting wps?
            if (len(self.wps_received) < self.num_wps_expected):
                #if we haven't already received this wp, write it to the GUI:
                if (m.seq not in self.wps_received.keys()):
                    self.gui_event_queue.put(MissionEditorEvent(
                        me_event.MEGE_SET_MISS_ITEM,
                        num=m.seq,command=m.command,param1=m.param1,
                        param2=m.param2,param3=m.param3,param4=m.param4,
                        lat=m.x,lon=m.y,alt=m.z,frame=m.frame))

                    self.wps_received[m.seq] = True

    def child_task(self, q, l, gq, gl, cw_sem):
        '''child process - this holds GUI elements'''
        mp_util.child_close_fds()

        from MAVProxy.modules.lib import wx_processguard
        from ..lib.wx_loader import wx
        from MAVProxy.modules.mavproxy_misseditor import missionEditorFrame

        self.app = wx.App(False)
        self.app.frame = missionEditorFrame.MissionEditorFrame(self,parent=None,id=wx.ID_ANY)

        self.app.frame.set_event_queue(q)
        self.app.frame.set_event_queue_lock(l)
        self.app.frame.set_gui_event_queue(gq)
        self.app.frame.set_gui_event_queue_lock(gl)
        self.app.frame.set_close_window_semaphore(cw_sem)

        self.app.SetExitOnFrameDelete(True)
        self.app.frame.Show()

        # start a thread to monitor the "close window" semaphore:
        class CloseWindowSemaphoreWatcher(threading.Thread):
            def __init__(self, task, sem):
                threading.Thread.__init__(self)
                self.task = task
                self.sem = sem
            def run(self):
                self.sem.acquire(True)
                self.task.app.ExitMainLoop()
        watcher_thread = CloseWindowSemaphoreWatcher(self, cw_sem)
        watcher_thread.start()

        self.app.MainLoop()
        # tell the watcher it is OK to quit:
        cw_sem.release()
        watcher_thread.join()

    def close(self):
        '''close the Mission Editor window'''
        self.time_to_quit = True
        self.close_window.release()
        if self.child.is_alive():
            self.child.join(1)

        self.child.terminate()

        self.mavlink_message_queue_handler.time_to_quit = True
        self.mavlink_message_queue_handler.join()

        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_TIME_TO_QUIT));
        self.event_queue_lock.release()

        self.needs_unloading = True

    def read_waypoints(self):
        self.module('wp').cmd_wp(['list'])

    def update_map_click_position(self, new_click_pos):
        self.gui_event_queue_lock.acquire()
        self.gui_event_queue.put(MissionEditorEvent(
            me_event.MEGE_SET_LAST_MAP_CLICK_POS,click_pos=new_click_pos))
        self.gui_event_queue_lock.release()

    def set_layout(self, layout):
        self.object_queue.put(layout)

def init(mpstate):
    '''initialise module'''
    return MissionEditorModule(mpstate)



#!/usr/bin/env python
'''miscellaneous commands'''

import time, math, sys
from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util


from os import kill
from signal import signal
from subprocess import PIPE, Popen

class RepeatCommand(object):
    '''repeated command object'''
    def __init__(self, period, cmd):
        self.period = period
        self.cmd = cmd
        self.event = mavutil.periodic_event(1.0/period)

    def __str__(self):
        return "Every %.1f seconds: %s" % (self.period, self.cmd)


def run_command(args, cwd = None, shell = False, timeout = None, env = None):
    '''
    Run a shell command with a timeout.
    See http://stackoverflow.com/questions/1191374/subprocess-with-timeout
    '''
    from subprocess import PIPE, Popen
    try:
        # py2
        from StringIO import StringIO
    except ImportError:
        # py3
        from io import StringIO
    import fcntl, os, signal
    p = Popen(args, shell = shell, cwd = cwd, stdout = PIPE, stderr = PIPE, env = env)
    tstart = time.time()
    buf = StringIO()

    # try to make it non-blocking
    try:
        fcntl.fcntl(p.stdout, fcntl.F_SETFL, fcntl.fcntl(p.stdout, fcntl.F_GETFL) | os.O_NONBLOCK)
    except Exception as ex:
        pass

    while True:
        time.sleep(0.1)
        retcode = p.poll()
        try:
            s = p.stdout.read()
            if sys.version_info.major >= 3:
                s = s.decode('utf-8')
            buf.write(s)
        except Exception as ex:
            pass
        if retcode is not None:
            break
        if timeout is not None and time.time() > tstart + timeout:
            print("timeout in process %u" % p.pid)
            try:
                os.kill(p.pid, signal.SIGKILL)
            except OSError:
                pass
            p.wait()
    return buf.getvalue()

class MiscModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(MiscModule, self).__init__(mpstate, "misc", "misc commands", public=True)
        self.add_command('alt', self.cmd_alt, "show altitude information")
        self.add_command('up', self.cmd_up, "adjust pitch trim by up to 5 degrees")
        self.add_command('reboot', self.cmd_reboot, "reboot autopilot")
        self.add_command('time', self.cmd_time, "show autopilot time")
        self.add_command('shell', self.cmd_shell, "run shell command")
        self.add_command('changealt', self.cmd_changealt, "change target altitude")
        self.add_command('changealt_abs', self.cmd_changealt_abs, "change target absolute altitude")
        self.add_command('land', self.cmd_land, "auto land")
        self.add_command('repeat', self.cmd_repeat, "repeat a command at regular intervals",
                         ["<add|remove|clear>"])
        self.add_command('version', self.cmd_version, "fetch autopilot version")
        self.add_command('capabilities', self.cmd_capabilities, "fetch autopilot capabilities")
        self.add_command('rcbind', self.cmd_rcbind, "bind RC receiver")
        self.add_command('led', self.cmd_led, "control board LED")
        self.add_command('oreoled', self.cmd_oreoled, "control OreoLEDs")
        self.add_command('playtune', self.cmd_playtune, "play tune remotely")
        self.add_command('devid', self.cmd_devid, "show device names from parameter IDs")
        self.add_command('gethome', self.cmd_gethome, "get HOME_POSITION")
        self.add_command('flashbootloader', self.cmd_flashbootloader, "flash bootloader (dangerous)")
        self.add_command('wipe_parameters', self.cmd_wipe_parameters, "wipe autopilot parameters")
        self.add_command('lockup_autopilot', self.cmd_lockup_autopilot, "lockup autopilot")
        self.add_command('corrupt_params', self.cmd_corrupt_param, "corrupt param storage")
        self.add_command('hardfault_autopilot', self.cmd_hardfault_autopilot, "hardfault autopilot")
        self.add_command('panic_autopilot', self.cmd_panic_autopilot, "panic autopilot")
        self.add_command('longloop_autopilot', self.cmd_longloop_autopilot, "cause long loop in autopilot")
        self.add_command('internalerror_autopilot', self.cmd_internalerror_autopilot, "cause internal error in autopilot")
        self.add_command('batreset', self.cmd_battery_reset, "reset battery remaining")
        self.add_command('setorigin', self.cmd_setorigin, "set global origin")
        self.add_command('magsetfield', self.cmd_magset_field, "set expected mag field by field")
        self.add_command('magresetofs', self.cmd_magreset_ofs, "reset offsets for all compasses")
        self.add_command('namedvaluefloat', self.cmd_namedvaluefloat, "send a NAMED_VALUE_FLOAT")
        self.add_command('scripting', self.cmd_scripting, "control onboard scripting", ["<stop|restart>"])
        self.add_command('formatsdcard', self.cmd_formatsdcard, "format SD card")
        self.add_command('canforward', self.cmd_canforward, "enable CAN forwarding")

        self.repeats = []

    def altitude_difference(self, pressure1, pressure2, ground_temp):
        '''calculate barometric altitude'''
        scaling = pressure2 / pressure1
        temp = ground_temp + 273.15
        return 153.8462 * temp * (1.0 - math.exp(0.190259 * math.log(scaling)))

    def qnh_estimate(self):
        '''estimate QNH pressure from GPS altitude and scaled pressure'''
        alt_gps = self.master.field('GPS_RAW_INT', 'alt', 0) * 0.001
        pressure2 = self.master.field('SCALED_PRESSURE', 'press_abs', 0)
        ground_temp = self.get_mav_param('GND_TEMP', 21)
        temp = ground_temp + 273.15
        pressure1 = pressure2 / math.exp(math.log(1.0 - (alt_gps / (153.8462 * temp))) / 0.190259)
        return pressure1

    def cmd_alt(self, args):
        '''show altitude'''
        print("Altitude:  %.1f" % self.status.altitude)
        qnh_pressure = self.get_mav_param('AFS_QNH_PRESSURE', None)
        if qnh_pressure is not None and qnh_pressure > 0:
            ground_temp = self.get_mav_param('GND_TEMP', 21)
            pressure = self.master.field('SCALED_PRESSURE', 'press_abs', 0)
            qnh_alt = self.altitude_difference(qnh_pressure, pressure, ground_temp)
            print("QNH Alt: %u meters %u feet for QNH pressure %.1f" % (qnh_alt, qnh_alt*3.2808, qnh_pressure))
        print("QNH Estimate: %.1f millibars" % self.qnh_estimate())

    def cmd_shell(self, args):
        '''shell command'''
        print(run_command(args, shell=False, timeout=3))

    def cmd_up(self, args):
        '''adjust TRIM_PITCH_CD up by 5 degrees'''
        if len(args) == 0:
            adjust = 5.0
        else:
            adjust = float(args[0])
        old_trim = self.get_mav_param('TRIM_PITCH_CD', None)
        if old_trim is None:
            print("Existing trim value unknown!")
            return
        new_trim = int(old_trim + (adjust*100))
        if math.fabs(new_trim - old_trim) > 1000:
            print("Adjustment by %d too large (from %d to %d)" % (adjust*100, old_trim, new_trim))
            return
        print("Adjusting TRIM_PITCH_CD from %d to %d" % (old_trim, new_trim))
        self.param_set('TRIM_PITCH_CD', new_trim)

    def cmd_reboot(self, args):
        '''reboot autopilot'''
        if len(args) > 0 and args[0] == 'bootloader':
            self.master.reboot_autopilot(True)
        else:
            self.master.reboot_autopilot()

    def cmd_wipe_parameters(self, args):
        self.master.mav.command_long_send(
            self.settings.target_system,
            self.settings.target_component,
            mavutil.mavlink.MAV_CMD_PREFLIGHT_STORAGE,
            0,
            2,
            0,
            0,
            0,
            0,
            0,
            0)

    def cmd_dosomethingreallynastyto_autopilot(self, args, description, code):
        '''helper function for the following commands which do unpleasant
        things to the autopilot'''
        if len(args) > 0 and args[0] == 'IREALLYMEANIT':
            print("Sending %s command" % description)
            self.master.mav.command_long_send(
                self.settings.target_system,
                self.settings.target_component,
                mavutil.mavlink.MAV_CMD_PREFLIGHT_REBOOT_SHUTDOWN, 0,
                42,
                24,
                71,
                code,
                0,
                0,
                0)
        else:
            print("Invalid %s command" % description)

    def cmd_lockup_autopilot(self, args):
        '''lockup autopilot for watchdog testing'''
        self.cmd_dosomethingreallynastyto_autopilot(args, 'lockup', 93)

    def cmd_hardfault_autopilot(self, args):
        '''lockup autopilot for watchdog testing'''
        self.cmd_dosomethingreallynastyto_autopilot(args, 'hardfault', 94)

    def cmd_panic_autopilot(self, args):
        '''get ArduPilot to call AP_HAL::panic()'''
        self.cmd_dosomethingreallynastyto_autopilot(args, 'panic', 95)

    def cmd_corrupt_param(self, args):
        '''corrupt parameter storage for backup testing'''
        self.cmd_dosomethingreallynastyto_autopilot(args, 'corruption', 96)

    def cmd_longloop_autopilot(self, args):
        '''Ask the autopilot to create a long loop'''
        self.cmd_dosomethingreallynastyto_autopilot(args, 'long-loop', 97)

    def cmd_internalerror_autopilot(self, args):
        '''Ask the autopilot to create an internal error'''
        self.cmd_dosomethingreallynastyto_autopilot(args, 'internal-error', 98)

    def cmd_battery_reset(self, args):
        '''reset battery remaining'''
        mask = -1
        remaining_pct = 100
        if len(args) > 0:
            mask = int(args[0])
        if len(args) > 1:
            remaining_pct = int(args[1])
        self.master.mav.command_long_send(self.settings.target_system, self.settings.target_component,
                                          mavutil.mavlink.MAV_CMD_BATTERY_RESET, 0,
                                              mask, remaining_pct, 0, 0, 0, 0, 0)

    def cmd_time(self, args):
        '''show autopilot time'''
        tusec = self.master.field('SYSTEM_TIME', 'time_unix_usec', 0)
        if tusec == 0:
            print("No SYSTEM_TIME time available")
            return
        print("%s (%s)\n" % (time.ctime(tusec * 1.0e-6), time.ctime()))

    def cmd_changealt(self, args):
        '''change target altitude'''
        if len(args) < 1:
            print("usage: changealt <relaltitude>")
            return
        relalt = float(args[0])
        self.master.mav.mission_item_send(self.settings.target_system,
                                          self.settings.target_component,
                                          0,
                                          3,
                                          mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                          3, 1, 0, 0, 0, 0,
                                          0, 0, relalt)
        print("Sent change altitude command for %.1f meters" % relalt)

    def cmd_changealt_abs(self, args):
        '''change target altitude'''
        if len(args) < 1:
            print("usage: changealt <relaltitude>")
            return
        absalt = float(args[0])
        self.master.mav.mission_item_send(self.settings.target_system,
                                          self.settings.target_component,
                                          0,
                                          0,
                                          mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                          3, 1, 0, 0, 0, 0,
                                          0, 0, absalt)
        print("Sent change altitude command for %.1f meters" % absalt)
        
    def cmd_land(self, args):
        '''auto land commands'''
        if len(args) < 1:
            self.master.mav.command_long_send(self.settings.target_system,
                                              0,
                                              mavutil.mavlink.MAV_CMD_DO_LAND_START,
                                              0, 0, 0, 0, 0, 0, 0, 0)
        elif args[0] == 'abort':
            self.master.mav.command_long_send(self.settings.target_system,
                                              0,
                                              mavutil.mavlink.MAV_CMD_DO_GO_AROUND,
                                              0, 0, 0, 0, 0, 0, 0, 0)
        else:
            print("Usage: land [abort]")

    def cmd_version(self, args):
        '''show version'''
        self.master.mav.autopilot_version_request_send(self.settings.target_system,
                                                       self.settings.target_component)

    def cmd_capabilities(self, args):
        '''show capabilities'''
        self.master.mav.command_long_send(self.settings.target_system,
                                          self.settings.target_component,
                                          mavutil.mavlink.MAV_CMD_REQUEST_AUTOPILOT_CAPABILITIES,
                                          0,
                                          1, 0, 0, 0, 0, 0, 0)
        
    def cmd_rcbind(self, args):
        '''start RC bind'''
        if len(args) < 1:
            print("Usage: rcbind <dsmmode>")
            return
        self.master.mav.command_long_send(self.settings.target_system,
                                          self.settings.target_component,
                                          mavutil.mavlink.MAV_CMD_START_RX_PAIR,
                                          0,
                                          float(args[0]), 0, 0, 0, 0, 0, 0)

    def cmd_gethome(self, args):
        '''get home position'''
        self.master.mav.command_long_send(self.settings.target_system,
                                          0,
                                          mavutil.mavlink.MAV_CMD_GET_HOME_POSITION,
                                          0, 0, 0, 0, 0, 0, 0, 0)

    def cmd_led(self, args):
        '''send LED pattern as override'''
        if len(args) < 3:
            print("Usage: led RED GREEN BLUE <RATE>")
            return
        pattern = [0] * 24
        pattern[0] = int(args[0])
        pattern[1] = int(args[1])
        pattern[2] = int(args[2])
        
        if len(args) == 4:
            plen = 4
            pattern[3] = int(args[3])
        else:
            plen = 3
            
        self.master.mav.led_control_send(self.settings.target_system,
                                         self.settings.target_component,
                                         0, 0, plen, pattern)

    def cmd_scripting(self, args):
        '''control onboard scripting'''
        if len(args) < 1:
            print("Usage: scripting <stop|restart>")
            return

        if args[0] == 'restart':
            cmd = mavutil.mavlink.SCRIPTING_CMD_STOP_AND_RESTART
        elif args[0] == 'stop':
            cmd = mavutil.mavlink.SCRIPTING_CMD_STOP
        else:
            print("Usage: scripting <stop|restart>")
            return

        # MAVProxy command to stop and re-start is: command_int 0 42701 0 0 3 0 0 0 0 0 0

        self.master.mav.command_int_send(
            self.settings.target_system, self.settings.target_component,
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT_INT,
            mavutil.mavlink.MAV_CMD_SCRIPTING,
            0, 0,
            cmd,
            0,0,0,0,0,0)

    def cmd_formatsdcard(self, args):
        '''format SD card'''
        self.master.mav.command_int_send(
            self.settings.target_system, self.settings.target_component,
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT_INT,
            mavutil.mavlink.MAV_CMD_STORAGE_FORMAT,
            0, 0,
            1, 1,
            0,0,0,0,0)
        
    def cmd_oreoled(self, args):
        '''send LED pattern as override, using OreoLED conventions'''
        if len(args) < 4:
            print("Usage: oreoled LEDNUM RED GREEN BLUE <RATE>")
            return
        lednum = int(args[0])
        pattern = [0] * 24
        pattern[0] = ord('R')
        pattern[1] = ord('G')
        pattern[2] = ord('B')
        pattern[3] = ord('0')
        pattern[4] = 0
        pattern[5] = int(args[1])
        pattern[6] = int(args[2])
        pattern[7] = int(args[3])
        
        self.master.mav.led_control_send(self.settings.target_system,
                                         self.settings.target_component,
                                         lednum, 255, 8, pattern)
        
    def cmd_flashbootloader(self, args):
        '''flash bootloader'''
        self.master.mav.command_long_send(self.settings.target_system,
                                          0,
                                          mavutil.mavlink.MAV_CMD_FLASH_BOOTLOADER,
                                              0, 0, 0, 0, 0, 290876, 0, 0)
        
    def cmd_playtune(self, args):
        '''send PLAY_TUNE message'''
        if len(args) < 1:
            print("Usage: playtune TUNE")
            return
        tune = args[0]
        str1 = tune[0:30]
        str2 = tune[30:]
        if sys.version_info.major >= 3 and not isinstance(str1, bytes):
            str1 = bytes(str1, "ascii")
        if sys.version_info.major >= 3 and not isinstance(str2, bytes):
            str2 = bytes(str2, "ascii")
        self.master.mav.play_tune_send(self.settings.target_system,
                                       self.settings.target_component,
                                       str1, str2)

    def cmd_repeat(self, args):
        '''repeat a command at regular intervals'''
        if len(args) == 0:
            if len(self.repeats) == 0:
                print("No repeats")
                return
            for i in range(len(self.repeats)):
                print("%u: %s" % (i, self.repeats[i]))
            return
        if args[0] == 'add':
            if len(args) < 3:
                print("Usage: repeat add PERIOD CMD")
                return
            self.repeats.append(RepeatCommand(float(args[1]), " ".join(args[2:])))
        elif args[0] == 'remove':
            if len(args) < 2:
                print("Usage: repeat remove INDEX")
                return
            i = int(args[1])
            if i < 0 or i >= len(self.repeats):
                print("Invalid index %d" % i)
                return
            self.repeats.pop(i)
            return
        elif args[0] == 'clean':
            self.repeats = []
        else:
            print("Usage: repeat <add|remove|clean>")

    def cmd_devid(self, args):
        '''decode device IDs from parameters'''
        for p in self.mav_param.keys():
            if p.startswith('COMPASS_DEV_ID') or p.startswith('COMPASS_PRIO') or (
                    p.startswith('COMPASS') and p.endswith('DEV_ID')):
                mp_util.decode_devid(self.mav_param[p], p)
            if p.startswith('INS_') and p.endswith('_ID'):
                mp_util.decode_devid(self.mav_param[p], p)
            if p.startswith('GND_BARO') and p.endswith('_ID'):
                mp_util.decode_devid(self.mav_param[p], p)
            if p.startswith('BARO') and p.endswith('_DEVID'):
                mp_util.decode_devid(self.mav_param[p], p)
            if p.startswith('ARSPD') and p.endswith('_DEVID'):
                mp_util.decode_devid(self.mav_param[p], p)

    def cmd_setorigin(self, args):
        '''set global origin'''
        if len(args) < 3:
            print("Usage: setorigin LAT(deg) LON(deg) ALT(m)")
            return
        lat = float(args[0])
        lon = float(args[1])
        alt = float(args[2])
        print("Setting origin to: ", lat, lon, alt)
        self.master.mav.set_gps_global_origin_send(
            self.settings.target_system,
            int(lat*10000000), # lat
            int(lon*10000000), # lon
            int(alt*1000)) # param7

    def cmd_magset_field(self, args):
        '''set compass offsets by field'''
        if len(args) < 3:
            print("Usage: magsetfield MagX MagY MagZ")
            return
        magX = int(args[0])
        magY = int(args[1])
        magZ = int(args[2])

        field1x = self.master.field('RAW_IMU', 'xmag', 0)
        field1y = self.master.field('RAW_IMU', 'ymag', 0)
        field1z = self.master.field('RAW_IMU', 'zmag', 0)

        field2x = self.master.field('SCALED_IMU2', 'xmag', 0)
        field2y = self.master.field('SCALED_IMU2', 'ymag', 0)
        field2z = self.master.field('SCALED_IMU2', 'zmag', 0)

        field3x = self.master.field('SCALED_IMU3', 'xmag', 0)
        field3y = self.master.field('SCALED_IMU3', 'ymag', 0)
        field3z = self.master.field('SCALED_IMU3', 'zmag', 0)

        self.param_set('COMPASS_OFS_X', magX - (field1x - self.get_mav_param('COMPASS_OFS_X', 0)))
        self.param_set('COMPASS_OFS_Y', magY - (field1y - self.get_mav_param('COMPASS_OFS_Y', 0)))
        self.param_set('COMPASS_OFS_Z', magZ - (field1z - self.get_mav_param('COMPASS_OFS_Z', 0)))

        self.param_set('COMPASS_OFS2_X', magX - (field2x - self.get_mav_param('COMPASS_OFS2_X', 0)))
        self.param_set('COMPASS_OFS2_Y', magY - (field2y - self.get_mav_param('COMPASS_OFS2_Y', 0)))
        self.param_set('COMPASS_OFS2_Z', magZ - (field2z - self.get_mav_param('COMPASS_OFS2_Z', 0)))

        self.param_set('COMPASS_OFS3_X', magX - (field3x - self.get_mav_param('COMPASS_OFS3_X', 0)))
        self.param_set('COMPASS_OFS3_Y', magY - (field3y - self.get_mav_param('COMPASS_OFS3_Y', 0)))
        self.param_set('COMPASS_OFS3_Z', magZ - (field3z - self.get_mav_param('COMPASS_OFS3_Z', 0)))

    def cmd_magreset_ofs(self, args):
        '''set compass offsets to all zero'''
        self.param_set('COMPASS_OFS_X', 0)
        self.param_set('COMPASS_OFS_Y', 0)
        self.param_set('COMPASS_OFS_Z', 0)
        self.param_set('COMPASS_DIA_X', 1)
        self.param_set('COMPASS_DIA_Y', 1)
        self.param_set('COMPASS_DIA_Z', 1)
        self.param_set('COMPASS_ODI_X', 0)
        self.param_set('COMPASS_ODI_Y', 0)
        self.param_set('COMPASS_ODI_Z', 0)

        self.param_set('COMPASS_OFS2_X', 0)
        self.param_set('COMPASS_OFS2_Y', 0)
        self.param_set('COMPASS_OFS2_Z', 0)
        self.param_set('COMPASS_DIA2_X', 1)
        self.param_set('COMPASS_DIA2_Y', 1)
        self.param_set('COMPASS_DIA2_Z', 1)
        self.param_set('COMPASS_ODI2_X', 0)
        self.param_set('COMPASS_ODI2_Y', 0)
        self.param_set('COMPASS_ODI2_Z', 0)

        self.param_set('COMPASS_OFS3_X', 0)
        self.param_set('COMPASS_OFS3_Y', 0)
        self.param_set('COMPASS_OFS3_Z', 0)
        self.param_set('COMPASS_DIA3_X', 1)
        self.param_set('COMPASS_DIA3_Y', 1)
        self.param_set('COMPASS_DIA3_Z', 1)
        self.param_set('COMPASS_ODI3_X', 0)
        self.param_set('COMPASS_ODI3_Y', 0)
        self.param_set('COMPASS_ODI3_Z', 0)

    def cmd_namedvaluefloat(self, args):
        '''send a NAMED_VALUE_FLOAT'''
        if len(args) < 2:
            print("Usage: namedvaluefloat NAME value")
            return
        tnow_ms = int((time.time() - self.mpstate.start_time_s)*1000)
        name = args[0]
        value = float(args[1])
        self.master.mav.named_value_float_send(tnow_ms, name.encode("utf-8"), value)

    def cmd_canforward(self, args):
        if len(args) < 1:
            print("Usage: canforward bus")
            return
        bus = int(args[0])
        self.master.mav.command_long_send(
            self.settings.target_system,
            self.settings.target_component,
            mavutil.mavlink.MAV_CMD_CAN_FORWARD,
            0,
            bus,
            0,
            0,
            0,
            0,
            0,
            0)
        
    def idle_task(self):
        '''called on idle'''
        for r in self.repeats:
            if r.event.trigger():
                self.mpstate.functions.process_stdin(r.cmd, immediate=True)

def init(mpstate):
    '''initialise module'''
    return MiscModule(mpstate)



#!/usr/bin/env python
'''log command handling'''

import time, os

from MAVProxy.modules.lib import mp_module

class LogModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(LogModule, self).__init__(mpstate, "log", "log transfer")
        self.add_command('log', self.cmd_log, "log file handling", ['<download|status|erase|resume|cancel|list>'])
        self.reset()

    def reset(self):
        self.download_set = set()
        self.download_file = None
        self.download_lognum = None
        self.download_filename = None
        self.download_start = None
        self.download_last_timestamp = None
        self.download_ofs = 0
        self.retries = 0
        self.entries = {}
        self.download_queue = []
        self.last_status = time.time()

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'LOG_ENTRY':
            self.handle_log_entry(m)
        elif m.get_type() == 'LOG_DATA':
            self.handle_log_data(m)

    def handle_log_entry(self, m):
        '''handling incoming log entry'''
        if m.time_utc == 0:
            tstring = ''
        else:
            tstring = time.ctime(m.time_utc)
        if m.num_logs == 0:
            print("No logs")
            return
        self.entries[m.id] = m
        print("Log %u  numLogs %u lastLog %u size %u %s" % (m.id, m.num_logs, m.last_log_num, m.size, tstring))


    def handle_log_data(self, m):
        '''handling incoming log data'''
        if self.download_file is None:
            return
        # lose some data
        # import random
        # if random.uniform(0,1) < 0.05:
        #    print('dropping ', str(m))
        #    return
        if m.ofs != self.download_ofs:
            self.download_file.seek(m.ofs)
            self.download_ofs = m.ofs
        if m.count != 0:
            s = bytearray(m.data[:m.count])
            self.download_file.write(s)
            self.download_set.add(m.ofs // 90)
            self.download_ofs += m.count
        self.download_last_timestamp = time.time()
        if m.count == 0 or (m.count < 90 and len(self.download_set) == 1 + (m.ofs // 90)):
            dt = time.time() - self.download_start
            self.download_file.close()
            size = os.path.getsize(self.download_filename)
            speed = size / (1000.0 * dt)
            status = "Finished downloading %s (%u bytes %u seconds, %.1f kbyte/sec %u retries)" % (self.download_filename,
                                                                                                   size,
                                                                                                   dt, speed,
                                                                                                   self.retries)
            self.console.set_status('LogDownload',status, row=4)
            print(status)
            self.download_file = None
            self.download_filename = None
            self.download_set = set()
            self.master.mav.log_request_end_send(self.target_system,
                                                 self.target_component)
            if len(self.download_queue):
                self.log_download_next()
        self.update_status()

    def handle_log_data_missing(self):
        '''handling missing incoming log data'''
        if len(self.download_set) == 0:
            return
        highest = max(self.download_set)
        diff = set(range(highest)).difference(self.download_set)
        if len(diff) == 0:
            self.master.mav.log_request_data_send(self.target_system,
                                                       self.target_component,
                                                       self.download_lognum, (1 + highest) * 90, 0xffffffff)
            self.retries += 1
        else:
            num_requests = 0
            while num_requests < 20:
                start = min(diff)
                diff.remove(start)
                end = start
                while end + 1 in diff:
                    end += 1
                    diff.remove(end)
                self.master.mav.log_request_data_send(self.target_system,
                                                           self.target_component,
                                                           self.download_lognum, start * 90, (end + 1 - start) * 90)
                num_requests += 1
                self.retries += 1
                if len(diff) == 0:
                    break


    def log_status(self, console=False):
        '''show download status'''
        if self.download_filename is None:
            print("No download")
            return
        dt = time.time() - self.download_start
        speed = os.path.getsize(self.download_filename) / (1000.0 * dt)
        m = self.entries.get(self.download_lognum, None)
        file_size = os.path.getsize(self.download_filename)
        if m is None:
            size = 0
            pct = 0
        elif m.size == 0:
            size = 0
            pct = 100
        else:
            size = m.size
            pct = (100.0*file_size)/size
        highest = 0
        if len(self.download_set):
            highest = max(self.download_set)
        diff = set(range(highest)).difference(self.download_set)
        status = "Downloading %s - %u/%u bytes %.1f%% %.1f kbyte/s (%u retries %u missing)" % (self.download_filename,
                                                                                            os.path.getsize(self.download_filename),
                                                                                        size,
                                                                                        pct,
                                                                                        speed,
                                                                                        self.retries,
                                                                                            len(diff))
        if console:
            self.console.set_status('LogDownload', status, row=4)
        else:
            print(status)

    def log_download_next(self):
        if len(self.download_queue) == 0:
            return
        latest = self.download_queue.pop()
        filename = self.default_log_filename(latest)
        if os.path.isfile(filename) and os.path.getsize(filename) == self.entries.get(latest).to_dict()["size"]:
            print("Skipping existing %s" % (filename))
            self.log_download_next()
        else:
            self.log_download(latest, filename)

    def log_download_all(self):
        if len(self.entries.keys()) == 0:
            print("Please use log list first")
            return
        self.download_queue = sorted(self.entries, key=lambda id: self.entries[id].time_utc)
        self.log_download_next()

    def log_download_range(self, first, last):
        self.download_queue = sorted(list(range(first,last+1)),reverse=True)
        print(self.download_queue)
        self.log_download_next()

    def log_download_from(self,fromnum = 0):
        if len(self.entries.keys()) == 0:
            print("Please use log list first")
            return
        self.download_queue = sorted(self.entries, key=lambda id: self.entries[id].time_utc)
        self.download_queue = self.download_queue[fromnum:len(self.download_queue)]
        self.log_download_next()

    def log_download(self, log_num, filename):
        '''download a log file'''
        print("Downloading log %u as %s" % (log_num, filename))
        self.download_lognum = log_num
        self.download_file = open(filename, "wb")
        self.master.mav.log_request_data_send(self.target_system,
                                                   self.target_component,
                                                   log_num, 0, 0xFFFFFFFF)
        self.download_filename = filename
        self.download_set = set()
        self.download_start = time.time()
        self.download_last_timestamp = time.time()
        self.download_ofs = 0
        self.retries = 0

    def default_log_filename(self, log_num):
        return "log%u.bin" % log_num

    def cmd_log(self, args):
        '''log commands'''
        usage = "usage: log <list|download|erase|resume|status|cancel>"
        if len(args) < 1:
            print(usage)
            return

        if args[0] == "status":
            self.log_status()
        elif args[0] == "list":
            print("Requesting log list")
            self.download_set = set()
            self.master.mav.log_request_list_send(self.target_system,
                                                       self.target_component,
                                                       0, 0xffff)

        elif args[0] == "erase":
            self.master.mav.log_erase_send(self.target_system,
                                                self.target_component)

        elif args[0] == "resume":
            self.master.mav.log_request_end_send(self.target_system,
                                                      self.target_component)

        elif args[0] == "cancel":
            if self.download_file is not None:
                self.download_file.close()
            self.reset()

        elif args[0] == "download":
            if len(args) < 2:
                print("usage: log download all | log download <lognumber> <filename> | log download from <lognumber>|log download range FIRST LAST")
                return
            if args[1] == 'all':
                self.log_download_all()
                return
            if args[1] == 'from':
                if len(args) < 2:
                    args[2] == 0
                self.log_download_from(int(args[2]))
                return
            if args[1] == 'range':
                if len(args) < 2:
                    print("Usage: log download range FIRST LAST")
                    return
                self.log_download_range(int(args[2]), int(args[3]))
                return
            if args[1] == 'latest':
                if len(self.entries.keys()) == 0:
                    print("Please use log list first")
                    return
                log_num = sorted(self.entries, key=lambda id: self.entries[id].time_utc)[-1]
            else:
                log_num = int(args[1])
            if len(args) > 2:
                filename = args[2]
            else:
                filename = self.default_log_filename(log_num)
            self.log_download(log_num, filename)
        else:
            print(usage)


    def update_status(self):
        '''update log download status in console'''
        now = time.time()
        if self.download_file is not None and now - self.last_status > 0.5:
            self.last_status = now
            self.log_status(True)
            
    def idle_task(self):
        '''handle missing log data'''
        if self.download_last_timestamp is not None and time.time() - self.download_last_timestamp > 0.7:
            self.download_last_timestamp = time.time()
            self.handle_log_data_missing()
        self.update_status()

def init(mpstate):
    '''initialise module'''
    return LogModule(mpstate)



'''
MAVProxy sailing dashboard utility classes
'''

import enum

class WindReference(enum.Enum):
    '''
    An enumeration for the wind reference (either RELATIVE or TRUE)
    '''

    RELATIVE = 1
    TRUE = 2

class SpeedUnit(enum.Enum):
    '''
    An enumeration for the units of speed: kph, mph, or knots
    '''

    KPH = 1
    MPH = 2
    KNOTS = 3

class WindAngleAndSpeed(object):
    '''
    Wind angle and speed attributes
    '''

    def __init__(self, reference, angle, speed, unit):
        self.wind_reference = reference
        self.wind_angle = angle
        self.wind_speed = speed
        self.wind_speed_unit = unit

class WaterSpeedAndHeading(object):
    '''
    Water speed and heading attributes
    '''

    def __init__(self, water_speed, heading_true):
        self.water_speed = water_speed
        self.heading_true = heading_true



#!/usr/bin/env python
'''
Message Rate viewer
Peter Barker, December 2018

Simply display message rates
'''

import time

from MAVProxy.modules.lib import mp_module
from pymavlink import mavutil

class messagerate(mp_module.MPModule):
    def __init__(self, mpstate):
        """Initialise module"""
        super(messagerate, self).__init__(mpstate, "messagerate", "")
        self.counts = {}
        self.buckets = []
        self.max_buckets = 5

        self.last_calc = time.time()
        self.add_command('messagerate',
                         self.cmd_messagerate,
                         "messagerate module",
                         ['status', 'reset', 'set', 'get'])

    def usage(self):
        '''show help on command line options'''
        return "Usage: messagerate <status | reset | set(msg)(rate) | get(msg)>"

    def cmd_messagerate(self, args):
        '''control behaviour of the module'''
        if len(args) == 0:
            print(self.usage())
        elif args[0] == "status":
            print(self.status())
        elif args[0] == "reset":
            self.reset()
        elif args[0] == "get":
            if len(args) != 2:
                print(self.usage())
                return
            message_name = args[1]
            mavlink_map = mavutil.mavlink.mavlink_map
            for msg_id in mavlink_map.keys():
                mav_cmd_name = mavlink_map[msg_id].name
                if mav_cmd_name == message_name:
                    self.master.mav.command_long_send(
                        self.settings.target_system,
                        self.settings.target_component,
                        mavutil.mavlink.MAV_CMD_GET_MESSAGE_INTERVAL,
                        0,
                        msg_id, 0, 0, 0, 0, 0, 0)
                    return
            print("Unknown message ID:%s" % message_name)
        elif args[0] == "set":
            if len(args) < 3:
                print(self.usage())
                return
            message_name = args[1]
            message_rate = float(args[2])
            priority = 0
            if len(args) > 3:
              priority = int(args[3])
            mavlink_map = mavutil.mavlink.mavlink_map
            for msg_id in mavlink_map.keys():
                mav_cmd_name = mavlink_map[msg_id].name
                if mav_cmd_name == message_name:
                    self.master.mav.command_long_send(
                        self.settings.target_system,
                        self.settings.target_component,
                        mavutil.mavlink.MAV_CMD_SET_MESSAGE_INTERVAL,
                        0,
                        msg_id, (int) (1E6/message_rate), priority, 0, 0, 0, 0)
                    return
            print("Unknown message ID:%s" % message_name)
            
        else:
            print(self.usage())

    def reset(self):
        '''reset rates'''
        self.counts = {}
        self.buckets = []
        self.last_calc = time.time()

    def status(self):
        '''returns rates'''
        counts = {}
        for bucket in self.buckets:
            for x in bucket:
                if not x in counts:
                    counts[x] = 0
                counts[x] += bucket[x]

        ret = ""
        mtypes = counts.keys()
        mtypes = sorted(mtypes)
        for mtype in mtypes:
            ret += "%s: %0.1f/s\n" % (mtype,
                                      counts[mtype]/float(len(self.buckets)))
        return ret

    def idle_task(self):
        '''called rapidly by mavproxy'''
        now = time.time()
        time_delta = now - self.last_calc
        if time_delta > 1:
            self.last_calc = now
            self.buckets.append(self.counts)
            self.counts = {}
            if len(self.buckets) > self.max_buckets:
                self.buckets = self.buckets[-self.max_buckets:]

    def mavlink_packet(self, m):
        '''handle mavlink packets'''
        mtype = m.get_type()
        if mtype not in self.counts:
            self.counts[mtype] = 0
        self.counts[mtype] += 1
        
        mavlink_map = mavutil.mavlink.mavlink_map
              
        if mtype ==  'MESSAGE_INTERVAL':
            if  m.message_id in mavlink_map:
                print("Msg:%s  rate:%0.2fHz" % (mavlink_map[m.message_id].name, (1E6/m.interval_us) ) )
            else:
                print("Msg ID:%s  rate:%0.2fHz" % (m.message_id, (1E6/m.interval_us) ) )


def init(mpstate):
    '''initialise module'''
    return messagerate(mpstate)



# Copyright (C) 2016  Intel Corporation. All rights reserved.
#
# This file is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
This module takes
https://github.com/ArduPilot/ardupilot/blob/master/libraries/AP_Math/AP_GeodesicGrid.h
and
https://github.com/ArduPilot/ardupilot/blob/master/libraries/AP_Math/AP_GeodesicGrid.cpp
as reference for defining the geodesic sections and implementing almost all
functions. Those files should be consulted for implementation details.
'''
import math

from pymavlink.rotmat import Matrix3, Vector3
# The golden number below was obtained from scipy.constants.golden. Let's use
# the literal value here as this is the only place that would require scipy
# module.
g = 1.618033988749895

_first_half = (
    (Vector3(-g, 1, 0), Vector3(-1, 0,-g), Vector3(-g,-1, 0)),
    (Vector3(-1, 0,-g), Vector3(-g,-1, 0), Vector3( 0,-g,-1)),
    (Vector3(-g,-1, 0), Vector3( 0,-g,-1), Vector3( 0,-g, 1)),
    (Vector3(-1, 0,-g), Vector3( 0,-g,-1), Vector3( 1, 0,-g)),
    (Vector3( 0,-g,-1), Vector3( 0,-g, 1), Vector3( g,-1, 0)),
    (Vector3( 0,-g,-1), Vector3( 1, 0,-g), Vector3( g,-1, 0)),
    (Vector3( g,-1, 0), Vector3( 1, 0,-g), Vector3( g, 1, 0)),
    (Vector3( 1, 0,-g), Vector3( g, 1, 0), Vector3( 0, g,-1)),
    (Vector3( 1, 0,-g), Vector3( 0, g,-1), Vector3(-1, 0,-g)),
    (Vector3( 0, g,-1), Vector3(-g, 1, 0), Vector3(-1, 0,-g)),
)
_second_half = tuple((-a, -b, -c) for a, b, c in _first_half)

triangles = _first_half + _second_half

radius = math.sqrt(1 + g**2)

# radius / (length of two vertices of an icosahedron triangle)
_midpoint_projection_scale = radius / (2 * g)

sections = ()
for a, b, c in triangles:
    ma = _midpoint_projection_scale * (a + b)
    mb = _midpoint_projection_scale * (b + c)
    mc = _midpoint_projection_scale * (c + a)

    sections += (
        (ma, mb, mc),
        ( a, ma, mc),
        (ma,  b, mb),
        (mc, mb,  c),
    )

del _first_half, _second_half

class _NeighborUmbrella:
    index_to_attr = ('x', 'y', 'z')
    def __init__(self, components, v0_c0, v1_c1, v2_c1, v4_c4, v0_c4):
        self.components = components
        self.v0_c0 = self.index_to_attr[v0_c0]
        self.v1_c1 = self.index_to_attr[v1_c1]
        self.v2_c1 = self.index_to_attr[v2_c1]
        self.v4_c4 = self.index_to_attr[v4_c4]
        self.v0_c4 = self.index_to_attr[v0_c4]

_neighbor_umbrellas = (
    _NeighborUmbrella(( 9,  8,  7, 12, 14), 1, 2, 0, 0, 2),
    _NeighborUmbrella(( 1,  2,  4,  5,  3), 0, 0, 2, 2, 0),
    _NeighborUmbrella((16, 15, 13, 18, 17), 2, 2, 0, 2, 1),
)

_inverses = (
    Matrix3(
        Vector3(-0.309017,  0.500000,  0.190983),
        Vector3( 0.000000,  0.000000, -0.618034),
        Vector3(-0.309017, -0.500000,  0.190983)),
    Matrix3(
        Vector3(-0.190983,  0.309017, -0.500000),
        Vector3(-0.500000, -0.190983,  0.309017),
        Vector3( 0.309017, -0.500000, -0.190983)
    ),
    Matrix3(
        Vector3(-0.618034,  0.000000,  0.000000),
        Vector3( 0.190983, -0.309017, -0.500000),
        Vector3( 0.190983, -0.309017,  0.500000)
    ),
    Matrix3(
        Vector3(-0.500000,  0.190983, -0.309017),
        Vector3( 0.000000, -0.618034,  0.000000),
        Vector3( 0.500000,  0.190983, -0.309017)
    ),
    Matrix3(
        Vector3(-0.190983, -0.309017, -0.500000),
        Vector3(-0.190983, -0.309017,  0.500000),
        Vector3( 0.618034,  0.000000,  0.000000)
    ),
    Matrix3(
        Vector3(-0.309017, -0.500000, -0.190983),
        Vector3( 0.190983,  0.309017, -0.500000),
        Vector3( 0.500000, -0.190983,  0.309017)
    ),
    Matrix3(
        Vector3( 0.309017, -0.500000,  0.190983),
        Vector3( 0.000000,  0.000000, -0.618034),
        Vector3( 0.309017,  0.500000,  0.190983)
    ),
    Matrix3(
        Vector3( 0.190983, -0.309017, -0.500000),
        Vector3( 0.500000,  0.190983,  0.309017),
        Vector3(-0.309017,  0.500000, -0.190983)
    ),
    Matrix3(
        Vector3( 0.500000, -0.190983, -0.309017),
        Vector3( 0.000000,  0.618034,  0.000000),
        Vector3(-0.500000, -0.190983, -0.309017)
    ),
    Matrix3(
        Vector3( 0.309017,  0.500000, -0.190983),
        Vector3(-0.500000,  0.190983,  0.309017),
        Vector3(-0.190983, -0.309017, -0.500000)
    ),
)

_mid_inverses = (
    Matrix3(
        Vector3(-0.000000,  1.000000, -0.618034),
        Vector3( 0.000000, -1.000000, -0.618034),
        Vector3(-0.618034,  0.000000,  1.000000)
    ),
    Matrix3(
        Vector3(-1.000000,  0.618034, -0.000000),
        Vector3(-0.000000, -1.000000,  0.618034),
        Vector3( 0.618034, -0.000000, -1.000000)
    ),
    Matrix3(
        Vector3(-0.618034, -0.000000, -1.000000),
        Vector3( 1.000000, -0.618034, -0.000000),
        Vector3(-0.618034,  0.000000,  1.000000)
    ),
    Matrix3(
        Vector3(-1.000000, -0.618034, -0.000000),
        Vector3( 1.000000, -0.618034,  0.000000),
        Vector3(-0.000000,  1.000000, -0.618034)
    ),
    Matrix3(
        Vector3(-1.000000, -0.618034,  0.000000),
        Vector3( 0.618034,  0.000000,  1.000000),
        Vector3( 0.618034,  0.000000, -1.000000)
    ),
    Matrix3(
        Vector3(-0.618034, -0.000000, -1.000000),
        Vector3( 1.000000,  0.618034, -0.000000),
        Vector3( 0.000000, -1.000000,  0.618034)
    ),
    Matrix3(
        Vector3( 0.000000, -1.000000, -0.618034),
        Vector3( 0.000000,  1.000000, -0.618034),
        Vector3( 0.618034, -0.000000,  1.000000)
    ),
    Matrix3(
        Vector3( 1.000000, -0.618034, -0.000000),
        Vector3( 0.000000,  1.000000,  0.618034),
        Vector3(-0.618034,  0.000000, -1.000000)
    ),
    Matrix3(
        Vector3( 1.000000,  0.618034, -0.000000),
        Vector3(-1.000000,  0.618034,  0.000000),
        Vector3( 0.000000, -1.000000, -0.618034)
    ),
    Matrix3(
        Vector3(-0.000000,  1.000000,  0.618034),
        Vector3(-1.000000, -0.618034, -0.000000),
        Vector3( 0.618034,  0.000000, -1.000000)
    ),
)

def get_section_hit(v):
    i = _triangle_index(v)
    if i < 0:
        return -1

    j =  _subtriangle_index(i, v)
    if j < 0:
        return -1

    return 4 * i + j

def _neighbor_umbrella_component(idx, comp_idx):
    if idx < 3:
        return _neighbor_umbrellas[idx].components[comp_idx]
    return (_neighbor_umbrellas[idx % 3].components[comp_idx] + 10) % 20

def _from_neighbor_umbrella(idx, v, u):
    # TODO: fix equality tests for floating point values (here and in the rest
    # of this file)
    if u.x == u.y:
        comp = _neighbor_umbrella_component(idx, 0)
        w = _inverses[comp % 10] * v
        if comp > 9:
            w = -w
        x0 = getattr(w, _neighbor_umbrellas[idx % 3].v0_c0)
        if x0 <= 0:
            return -1
        return comp

    if u.y > u.x:
        comp = _neighbor_umbrella_component(idx, 1)
        w = _inverses[comp % 10] * v
        if comp > 9:
            w = -w

        x1 = getattr(w, _neighbor_umbrellas[idx % 3].v1_c1)
        x2 = getattr(w, _neighbor_umbrellas[idx % 3].v2_c1)

        if x1 == 0:
            return -1
        elif x1 < 0:
            return _neighbor_umbrella_component(idx, 2)

        if x2 == 0:
            return -1
        elif x2 < 0:
            return _neighbor_umbrella_component(idx, 0)

        return comp
    else:
        comp = _neighbor_umbrella_component(idx, 4)
        w = _inverses[comp % 10] * v
        if comp > 9:
            w = -w
        x4 = getattr(w, _neighbor_umbrellas[idx % 3].v4_c4)
        x0 = getattr(w, _neighbor_umbrellas[idx % 3].v0_c4)

        if x4 == 0:
            return -1
        elif x4 < 0:
            return _neighbor_umbrella_component(idx, 0)

        if x0 == 0:
            return -1
        elif x0 < 0:
            return _neighbor_umbrella_component(idx, 3)

        return comp

def _triangle_index(v):
    w = _inverses[0] * v
    zero_count = 0
    balance = 0
    umbrella = -1

    if w.x == 0:
        zero_count += 1
    elif w.x > 0:
        balance += 1
    else:
        balance -= 1

    if w.y == 0:
        zero_count += 1
    elif w.y > 0:
        balance += 1
    else:
        balance -= 1

    if w.z == 0:
        zero_count += 1
    elif w.z > 0:
        balance += 1
    else:
        balance -= 1

    if balance == 3:
        return 0
    elif balance == -3:
        return 10
    elif balance == 2:
        return -1
    elif balance == -2:
        return -1
    elif balance == 1:
        if zero_count == 2:
            return -1

        if w.x < 0:
            umbrella = 1
        elif w.y < 0:
            umbrella = 2
        else:
            umbrella = 0
    elif balance == -1:
        if zero_count == 2:
            return -1

        if w.x > 0:
            umbrella = 4
        elif w.y > 0:
            umbrella = 5
        else:
            umbrella = 3

        w = -w
    elif balance == 0:
        if zero_count == 3:
            return -1

        if w.x < 0:
            umbrella = 1
        elif w.y < 0:
            umbrella = 2
        else:
            umbrella = 0

    m = umbrella % 3

    if m == 0:
        w.z = -w.z
    elif m == 1:
        w.x, w.y, w.z = w.y, w.z, -w.x
    elif m == 2:
        w.x, w.y, w.z = w.z, w.x, -w.y

    return _from_neighbor_umbrella(umbrella, v, w, inclusive)

def _subtriangle_index(triangle_index, v):
    w = _mid_inverses[triangle_index % 10] * v
    if triangle_index > 9:
        w = -w

    if 0 in (w.x, w.y, w.z):
        return -1

    if w.x < 0:
        return 3
    if w.y < 0:
        return 1
    if w.z < 0:
        return 2

    return 0



'''
setup a listening TCP socket for forwarding u-center connections to a GPS
via MAVLink serial-control
'''

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from pymavlink import mavutil
import socket
import time
import errno

class UcenterModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(UcenterModule, self).__init__(mpstate, "ucenter", "ucenter forwarding")

        self.add_command('ucenter', self.cmd_ucenter, "ucenter control",
                         ["<start|stop|restart>","set (UCENTERSETTING)"])

        self.ucenter_settings = mp_settings.MPSettings([("port", int, 2001),
                                                        ('devnum', int, 2),
                                                        ('baudrate', int, 115200),
                                                        ('delay', float, 0.01),
                                                        ('debug', int, 0)])
        self.add_completion_function('(UCENTERSETTING)',
                                     self.ucenter_settings.completion)
        self.listen_sock = None
        self.sock = None
        self.last_write = time.time()
        self.last_baudrate = 0
        self.last_devnum = -1

    def cmd_ucenter(self, args):
        '''ucenter command parser'''
        usage = "usage: ucenter <set|start|restart|stop>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.ucenter_settings.command(args[1:])
        elif args[0] == "start":
            self.start_listener()
        elif args[0] == "stop":
            self.stop_listener()
        elif args[0] == "restart":
            self.stop_listener()
            self.start_listener()
        else:
            print(usage)

    def start_listener(self):
        '''start listening for packets'''
        if self.listen_sock is not None:
            self.listen_sock.close()
        if self.sock is not None:
            self.sock.close()
        self.listen_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        self.listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.listen_sock.bind(('', self.ucenter_settings.port))
        self.listen_sock.setblocking(False)
        self.listen_sock.listen(1)
        print("ucenter listening on port %u" % self.ucenter_settings.port)

    def stop_listener(self):
        '''stop listening for packets'''
        if self.sock is not None:
            self.sock.close()
            self.sock = None
        if self.listen_sock is not None:
            self.listen_sock.close()
            self.listen_sock = None
        self.last_baudrate = 0

    def debug(self, s):
        '''debug write'''
        if self.ucenter_settings.debug <= 0:
            return
        print(s)

    def write(self, b):
        '''write some bytes to remove port'''
        if len(b) > 0:
            self.debug("sending '%s' (0x%02x) of len %u" % (b, ord(b[0]), len(b)))
        elif self.ucenter_settings.debug > 1:
            self.debug("sending empty request")
        while True:
            # note that we send a single empty buffer on len(b)==0
            n = len(b)
            if n > 70:
                n = 70
            buf = [ord(x) for x in b[:n]]
            buf.extend([0]*(70-len(buf)))

            if self.last_baudrate != self.ucenter_settings.baudrate or self.last_devnum != self.ucenter_settings.devnum:
                baudrate = self.ucenter_settings.baudrate
                self.last_baudrate = baudrate
                self.last_devnum = self.ucenter_settings.devnum
                print("ucenter requesting baudrate %u" % baudrate)
            else:
                baudrate = 0

            self.master.mav.serial_control_send(self.ucenter_settings.devnum,
                                                mavutil.mavlink.SERIAL_CONTROL_FLAG_EXCLUSIVE |
                                                mavutil.mavlink.SERIAL_CONTROL_FLAG_RESPOND,
                                                0,
                                                baudrate,
                                                n,
                                                buf)
            b = b[n:]
            self.last_write = time.time()
            if len(b) == 0:
                break

    def idle_task(self):
        '''called on idle'''
        if self.sock is None and self.listen_sock is None:
            return
        if self.sock is None:
            try:
                conn_sock, addr = self.listen_sock.accept()
            except Exception as e:
                if e.errno not in [ errno.EAGAIN, errno.EWOULDBLOCK ]:
                    print("ucenter listen fail")
                    self.stop_listener()
                    return
                return
            self.sock = conn_sock
            self.sock.setblocking(False)
            print("ucenter connection from %s" % str(addr))

        now = time.time()

        try:
            pkt = self.sock.recv(1000)
        except socket.error as e:
            if e.errno not in [ errno.EAGAIN, errno.EWOULDBLOCK ]:
                print("ucenter connection closed")
                self.sock.close()
                self.sock = None
                return
            # send empty packet if idle
            pkt = ''
        if len(pkt) == 0 and now - self.last_write < self.ucenter_settings.delay:
            return
        self.write(pkt)

    def mavlink_packet(self, m):
        '''process SERIAL_CONTROL packets'''
        if m.get_type() != 'SERIAL_CONTROL':
            return
        if self.sock is None:
            return
        if m.count == 0:
            return
        data = m.data
        data = m.data[:m.count]
        buf = ''.join(str(chr(x)) for x in data)
        self.debug("got reply len %u" % len(buf))
        try:
            self.sock.send(buf)
        except socket.error as e:
            if e.errno not in [ errno.EAGAIN, errno.EWOULDBLOCK ]:
                print("ucenter connection write error")
                self.sock.close()
                self.sock = None


def init(mpstate):
    '''initialise module'''
    return UcenterModule(mpstate)
